/*************************************************************
*
* Description:
*	Test program  for pC/OS and pC/TFS on AT91SAM7S...
*
* Author[s]:
*	Frank Goetze
**************************************************************/
#include "../pC_OS/inc/pC_OS.h"
#include "../pC_TFS/inc/pC_TFS.h"

#ifdef _OS_AT91SAM7Sxx_
#define REG8   (volatile U08 *)
#define REG16  (volatile U16 *)
#define REG32  (volatile U32 *)

typedef volatile U32 AT91_REG;              // all registers are correct 32bit aligned
typedef struct _AT91S_A32 {
    AT91_REG     s[32];                     // Source Mode Register & Source Vector Register
} *AT91S_A32;
typedef struct _AT91S_A04 {
    AT91_REG     s[4];                      // Endpoint Control and Status Register & Endpoint FIFO Data Register
} *AT91S_A04;

#define OSTimer_INT_PRIO      2                                       // INT prio of SYS-TC
#define AT91S_PIT_BASE        (0xFFFFFD30)
#define AT91S_PIT_PIMR        (*(REG32 (AT91S_PIT_BASE + 0x00)))      // Period Interval Mode Register
#define AT91S_PIT_PISR        (*(REG32 (AT91S_PIT_BASE + 0x04)))      // Period Interval Status Register
#define AT91S_PIT_PIVR        (*(REG32 (AT91S_PIT_BASE + 0x08)))      // Period Interval Value Register
#define AT91S_PIT_PIIR        (*(REG32 (AT91S_PIT_BASE + 0x0C)))      // Period Interval Image Register
#define AT91S_PIT_PITEN       ((U32) 0x1 << 24)                       // (PITC) Periodic Interval Timer Enabled
#define AT91S_PIT_PITIEN      ((U32) 0x1 << 25)                       // (PITC) Periodic Interval Timer Interrupt Enable
#define AT91S_AIC_BASE        (0xFFFFF000)
#define AT91S_AIC_SMR         ((AT91S_A32)(AT91S_AIC_BASE + 0x00))    // [32] Source Mode Register
#define AT91S_AIC_SVR         ((AT91S_A32)(AT91S_AIC_BASE + 0x80))    // [32] Source Vector Register
#define AT91S_AIC_IECR        (*(REG32 (AT91S_AIC_BASE + 0x120)))     // Interrupt Enable Command Register
#define AT91S_AIC_IDCR        (*(REG32 (AT91S_AIC_BASE + 0x124)))     // Interrupt Disable Command Register
#define AT91S_AIC_ICCR        (*(REG32 (AT91S_AIC_BASE + 0x128)))     // Interrupt Clear Command Register
#define AT91S_ID_SYS          1U                                      // System-Peripheral ID
#define AT91S_ID_PIO          2U                                      // Parallel IO Port ID
#define AT91S_AIC_SRCTYPE_INT_LEVEL_SENSITIVE      0x00000000U        // (AIC) Internal Sources Code Label Level Sensitive
#define AT91S_PIOA_BASE       (0xFFFFF400U)                           // only PIO-A in hardware
#define AT91S_PIO_PER         (*(REG32 (AT91S_PIOA_BASE + 0x00)))     // PIO Enable Register
#define AT91S_PIO_OER         (*(REG32 (AT91S_PIOA_BASE + 0x10)))     // Output Enable Register
#define AT91S_PIO_SODR        (*(REG32 (AT91S_PIOA_BASE + 0x30)))     // Set Output Data Register
#define AT91S_PIO_CODR        (*(REG32 (AT91S_PIOA_BASE + 0x34)))     // Clear Output Data Register
#define AT91S_PIO_OWER        (*(REG32 (AT91S_PIOA_BASE + 0xA0)))     // Output Write Enable Register
#define AT91S_PIO_PA00        (1U << 0)                               // Pin Controlled by pin PA00 (for AT91SAM7S-EX)
#define AT91S_PIO_PA01        (1U << 1)                               // Pin Controlled by pin PA01
#define AT91S_PIO_PA02        (1U << 2)                               // Pin Controlled by pin PA02
#define AT91S_PIO_PA11        (1U << 11)                              // Pin Controlled by pin PA11 (for AT91SAM7Sxx-DevMaster)
#define AT91S_PMC_BASE        (0xFFFFFC00U)
#define AT91S_PMC_PCER        (*(REG32 (AT91S_PMC_BASE + 0x10)))      // Peripheral Clock Enable Register

#define AT91S_PIO_LEDc        AT91S_PIO_PA11                          // Pin used for the control-LED (for AT91SAM7Sxx)
//#define AT91S_PIO_LEDc        AT91S_PIO_PA01                          // Pin used for the control-LED (for AT91SAM7S-EK / AT91SAM7SE-EK)
#endif


U08 LEDc  = 0xFE;
U08 MEMORY[1024];

#define OS_SYSTEM_TICKS_DIV   (OS_TIMER_TICKS_PER_SEC / OS_SYSTEM_TICKS_PER_SEC)

#define	TASK_STK_SIZE        512

OS_STK_TYPE Stk1[TASK_STK_SIZE];
OS_STK_TYPE Stk2[TASK_STK_SIZE];
OS_STK_TYPE Stk3[TASK_STK_SIZE];
OS_STK_TYPE Stk4[TASK_STK_SIZE];
OS_STK_TYPE Stk5[TASK_STK_SIZE];

U32 OS_SystemTicksCnt;

OS_MUX Mutex1;
OS_EVG Events1;
#if OS_TMR_EN
OS_SEM LedSem;
OS_TMR LedTimer;
OS_TMR ROnceTimer1;
OS_TMR ROnceTimer2;
#endif

void  Task1(void *Tdata);
void  Task2(void *Tdata);
void  Task3(void *Tdata);
void  Task4(void *Tdata);
void  Task5(void *Tdata);


int main(void)
{
    OS_Init();
    
    OS_MemoryInit((OS_MEM OS_HUGE *)MEMORY, 1024);
#if OS_STK_CHECK_EN
    OS_TaskCreate(Task1, NULL, (void *)&Stk1[TASK_STK_SIZE], &Stk1, 7);
    OS_TaskCreate(Task2, NULL, (void *)&Stk2[TASK_STK_SIZE], &Stk2, 9);
    OS_TaskCreate(Task3, NULL, (void *)&Stk3[TASK_STK_SIZE], &Stk3, 8);
    OS_TaskCreate(Task4, NULL, (void *)&Stk4[TASK_STK_SIZE], &Stk4, 12);
    OS_TaskCreate(Task5, NULL, (void *)&Stk5[TASK_STK_SIZE], &Stk5, 13);
#else
    OS_TaskCreate(Task1, NULL, (void *)&Stk1[TASK_STK_SIZE], 7);
    OS_TaskCreate(Task2, NULL, (void *)&Stk2[TASK_STK_SIZE], 9);
    OS_TaskCreate(Task3, NULL, (void *)&Stk3[TASK_STK_SIZE], 8);
    OS_TaskCreate(Task4, NULL, (void *)&Stk4[TASK_STK_SIZE], 12);
    OS_TaskCreate(Task5, NULL, (void *)&Stk5[TASK_STK_SIZE], 13);
#endif
    OS_MutexCreate(&Mutex1, 6);
    OS_EvgInit(&Events1);
#if OS_TMR_EN
    OS_SemInit(&LedSem, 0);
#endif

    TFS_Init();                        // init TiniFileSystem
    OS_Start();

    while(1);
}

void OSTickISR(void)
{
    volatile U32 status;

#ifdef _OS_AT91SAM7Sxx_
    status = AT91S_PIT_PIVR;                               // ack the interrupt
    AT91S_AIC_ICCR  = AT91S_ID_SYS;                        // clear interrupt
#else
 #error "unkown processor !"
#endif
#if OS_TMR_EN
    OSTimerTick();                                         // call pC/OS-function (timers)
#endif
    if((++OS_SystemTicksCnt) == OS_SYSTEM_TICKS_DIV) {
        OS_SystemTicksCnt = 0;
        OSSystemTick();                                    // call pC/OS-function (delays / wait)
    }
}

void AppTickInit(void)
{
    U32  counts;

    OS_SystemTicksCnt = 0;

#ifdef _OS_AT91SAM7Sxx_
    AT91S_PMC_PCER = (1 << AT91S_ID_PIO);                  // enable clock on PIO-mode
    AT91S_PIO_PER  = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Set in PIO mode
    AT91S_PIO_OER  = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Configure in Output
    AT91S_PIO_OWER = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // allow writing to this pins
    AT91S_PIO_CODR = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Clear pin to set the control-LED on (test)
    AT91S_PIO_SODR = AT91S_PIO_LEDc | AT91S_PIO_PA02;      // Set pin to clear the control-LED (test)

    AT91S_AIC_IDCR = (1 << AT91S_ID_SYS);                  // disable the interrupt
    AT91S_AIC_SVR->s[AT91S_ID_SYS] = (U32)(OSTickISR);     // set the ISR-function
    AT91S_AIC_SMR->s[AT91S_ID_SYS] = AT91S_AIC_SRCTYPE_INT_LEVEL_SENSITIVE | OSTimer_INT_PRIO  ;
    counts = (48000000 / 16 / OS_TIMER_TICKS_PER_SEC) - 1; // calculate counter for 48MHz using PLL
    AT91S_PIT_PIMR = AT91S_PIT_PITEN | AT91S_PIT_PITIEN | counts;
    AT91S_AIC_ICCR = (1 << AT91S_ID_SYS);                  // clear the interrupt
    AT91S_AIC_IECR = (1 << AT91S_ID_SYS);                  // enable the interrupt
#else
 #error "unkown processor !"
#endif
}

void BSP_toggle_LED(void)
{
    LEDc ^= 0x01;                                      // Toggle LED
    if(LEDc &0x01) {
        AT91S_PIO_CODR = AT91S_PIO_LEDc;               // Clear pin to set the control-LED on
        AT91S_PIO_SODR = AT91S_PIO_PA02;
    } else {
        AT91S_PIO_SODR = AT91S_PIO_LEDc;               // Set pin to clear the control-LED
        AT91S_PIO_CODR = AT91S_PIO_PA02;
    }
}

void Task1(void *Tdata)
{
    U08 returnOk;
    U32 event, i;
    U08 OS_HUGE *ptr;

    if(Tdata != NULL) while(1);
    AppTickInit();
    ptr=NULL;
    returnOk=OS_MemAlloc((void OS_HUGE *)(&ptr), 30);      // memory-manager test
    returnOk=OS_MemFree((OS_MEM OS_HUGE *)ptr);
    while(1) {
        i=0;
        while(i<2500) i++;                                 // no Task-switch to MUTEX
        returnOk=OS_MutexPend(&Mutex1, OS_SUSPEND);        // first ok. - secound suspending
        if(!returnOk) {
            OS_TimeDly(5);
            returnOk=OS_MutexPendAbbort(&Mutex1);
            OS_TimeDly(5);
            returnOk=OS_MutexPost(&Mutex1);
            OS_TimeDly(5);
            event = 0x10;
            returnOk=OS_EvgPend(&Events1, &event, OS_EVG_OR_C, OS_SUSPEND);
        }
    }
}

void Task2(void *Tdata)
{
    U08 returnOk;
    U32 i;

    Tdata = Tdata;
    while(1) {
        returnOk=OS_MutexPend(&Mutex1, OS_SUSPEND);        // suspending
        if(!returnOk) {
            i=0;
            while(i<2500) i++;                             // no Taskswitch to Task1 or Task3
            returnOk=OS_MutexPost(&Mutex1);
            OS_TimeDly(5);
            OS_EvgPost(&Events1, 0x10, OS_EVG_OR);
        }
    }
}

#if OS_TMR_EN
void ROnceTimer1CB(void *tdata)
{
    tdata = tdata;
    OS_TimerStop(&LedTimer);                  // stope LED-timer
}

void ROnceTimer2CB(void *tdata)
{
    tdata = tdata;
    OS_TimerStart(&LedTimer, 0);              // restart LED-timer with org time
    OS_TimerCreate(&ROnceTimer1, 500, ROnceTimer1CB, NULL, OS_TMR_ENABLE | OS_TMR_CLR);
    OS_TimerStart(&ROnceTimer2, 0);           // start both run-once (one auto-clear) timer again
}

void LedTimerCB(void *tdata)
{
    tdata = tdata;
    OS_SemPost(&LedSem);
}
#endif

void Task3(void *Tdata)
{
    Tdata = Tdata;                                         // Configure the LED
    
#if OS_TMR_EN
    OS_TimerCreate(&ROnceTimer1, 500, ROnceTimer1CB, NULL, OS_TMR_ENABLE | OS_TMR_CLR);
    OS_TimerCreate(&ROnceTimer2, 1000, ROnceTimer2CB, NULL, OS_TMR_ENABLE | OS_TMR_RONCE);
    OS_TimerCreate(&LedTimer, 30, LedTimerCB, NULL, OS_TMR_ENABLE | OS_TMR_CYCL);
#endif    
    while(1) {
#if OS_TMR_EN
        OS_SemPend(&LedSem, OS_SUSPEND);                   // wait on timer-triggert signal
#else
        OS_TimeDly(3);                                     // test timer-wakeup
#endif
        BSP_toggle_LED();                                      // Toggle LED
    }
}

//------------------------------- TFS user tasks ------------------------------------

TFS_USER TFS_T1, TFS_T2;

U08 Fill[]="______________________________________________";
U08 FillE[1074];
U08 Text[]="TFS_Test_File R/W";
void Task4(void *Tdata)
{
    U08            returnOk;
    U08            Readed[20];
    TFS_HANDLE     fhandle ,t;
    TFS_LONG       written;

    Tdata = Tdata;

    returnOk = TFS_Format();
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_BecomeUser(&TFS_T1);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_ChangeDir("/");
    fhandle =TFS_OpenFile("testE1", TFS_ACC_RW);
    if(!(fhandle))
        returnOk=TFS_GetErrNo();
    returnOk=TFS_CreateFile("testE1", TFS_ATTR_RW, 50);
    fhandle =TFS_OpenFile("testE1", TFS_ACC_RW);
    if(!(fhandle))
        returnOk=TFS_GetErrNo();
    written=TFS_WriteFileE(FillE, 1045, fhandle);
    if(written != 1045)
        returnOk=TFS_GetErrNo();
    returnOk=TFS_CreateFile("testE2", TFS_ATTR_RW, 50);
    t       =TFS_OpenFile("testE2", TFS_ACC_RW);
    if(!(t))
        returnOk=TFS_GetErrNo();
    written=TFS_WriteFileE(FillE, 1074, t);
    if(written != 1074)
        returnOk=TFS_GetErrNo();
    returnOk=TFS_CloseFile(fhandle);
    returnOk=TFS_CloseFile(t);
    returnOk=TFS_RemoveFile("testE1");
    returnOk=TFS_RemoveFile("testE2");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateDir("Dir1");
    if(returnOk==TFS_NO_ERR) {
        returnOk=TFS_CreateFile("Dir1/File11", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File12", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File13", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File14", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File15", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File16", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File17", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File18", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File19", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File20", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File21", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File22", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File23", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File24", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File25", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File26", TFS_ATTR_SYS, 100);
        returnOk=TFS_CreateFile("Dir1/File27", TFS_ATTR_SYS, 100);
    }
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("File1", TFS_ATTR_RO | TFS_ATTR_SYS, 100);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateDir("/Dir2");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("Dir2/File21", 0, 10);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("/Dir2/File22", TFS_ATTR_SYS, 300);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_ChangeDir("Dir2");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateDir("Dir22");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateDir("../Dir2/Dir21");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_RemoveDir("/Dir2/Dir22");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_ChangeDir("Dir21");
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("/Dir2/Dir21/File211", 0, 200);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("/Dir2/Dir21/RW_test", TFS_ATTR_SYS | TFS_ATTR_RO, 50);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("/Dir2/Dir21/RW_test2", TFS_ATTR_SYS, 400);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("/Dir2/Dir21/File212", TFS_ATTR_SYS, 150);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_ChangeDir("..");
    if(returnOk==TFS_NO_ERR) {
        returnOk=TFS_RemoveFile("/Dir1/File11");
        returnOk=TFS_RemoveFile("/Dir1/File13");
        returnOk=TFS_RemoveFile("/Dir1/File15");
        returnOk=TFS_RemoveFile("/Dir1/File17");
        returnOk=TFS_RemoveFile("/Dir1/File19");
        returnOk=TFS_RemoveFile("/Dir1/File21");
        returnOk=TFS_RemoveFile("/Dir1/File23");
        returnOk=TFS_RemoveFile("/Dir1/File25");
        returnOk=TFS_RemoveFile("/Dir1/File27");
    }
    OS_TimeDly(5);
    if(returnOk==TFS_NO_ERR)
        returnOk=TFS_CreateFile("../File2", TFS_ATTR_SYS, 500);
    if(returnOk==TFS_NO_ERR) {
        returnOk=TFS_AttribFile("Dir21/RW_test", TFS_ATTR_SYS);
        if(returnOk==TFS_NO_ERR)
            returnOk=TFS_ResizeFile("Dir21/RW_test", 400);
        if (returnOk==TFS_NO_ERR) {
            fhandle=TFS_OpenFile("/Dir2/Dir21/RW_test", TFS_ACC_RW);
            if(!(fhandle))
                returnOk=TFS_GetErrNo();
        }
        if(returnOk==TFS_NO_ERR) {
            written=TFS_WriteFile(Fill , sizeof(Fill), fhandle);
            if(written != sizeof(Fill))
                returnOk=TFS_GetErrNo();
            if(returnOk==TFS_NO_ERR)
                returnOk=TFS_SeekFile(45, fhandle);
            if(returnOk==TFS_NO_ERR) {
                written=TFS_WriteFile(Text , sizeof(Text), fhandle);
                if(written != sizeof(Text))
                    returnOk=TFS_GetErrNo();
                if(returnOk==TFS_NO_ERR) {
                    returnOk=TFS_SeekFile(45, fhandle);
                    if(returnOk==TFS_NO_ERR) {
                        written=TFS_ReadFile((U08 OS_HUGE *)Readed, sizeof(Text), fhandle);
                        if(written != sizeof(Text))
                            returnOk=TFS_GetErrNo();
                    }
                }
            }
            returnOk=TFS_CloseFile(fhandle);
            if(returnOk==TFS_NO_ERR)
                returnOk=TFS_RenameFile("/Dir2/Dir21/RW_test","config");
        }
    }
    while(1) {
        fhandle=TFS_OpenFile("/Dir2/Dir21/config", TFS_ACC_RW);
        if(fhandle) {
            returnOk=TFS_SeekFile(45, fhandle);
            if(returnOk==TFS_NO_ERR) {
                written=TFS_WriteFile(Text , sizeof(Text), fhandle);
                if(written != sizeof(Text))
                    returnOk=TFS_GetErrNo();
                if(returnOk==TFS_NO_ERR) {
                    returnOk=TFS_SeekFile(45, fhandle);
                    if(returnOk==TFS_NO_ERR) {
                        written=TFS_ReadFile((U08 OS_HUGE *)Readed, sizeof(Text), fhandle);
                        if(written != sizeof(Text))
                            returnOk=TFS_GetErrNo();
                    }
                }
            }
            returnOk=TFS_CloseFile(fhandle);
            BSP_toggle_LED();                                      // Toggle LED
        } else {
            returnOk=TFS_GetErrNo();
            if(returnOk == TFS_NO_ERR) 
                while(1);
        }
        OS_TimeDly(4);
    }
//lint -unreachable
//    returnOk=TFS_CloseUser();
}

void Task5(void *Tdata)
{
    U08            returnOk;
    U08            Readed[20];
    TFS_HANDLE     fhandle;
    TFS_LONG       written;

    Tdata = Tdata;

    returnOk=OS_TimeDlyResume(12);
    while(1) {
        returnOk=TFS_BecomeUserSubROOT(&TFS_T2, "/Dir2/Dir21/");
        if(returnOk==TFS_NO_ERR) {
            fhandle=TFS_OpenFile("config", TFS_ACC_RW);
            if(fhandle) {
                returnOk=TFS_SeekFile(45, fhandle);
                if(returnOk==TFS_NO_ERR) {
                    written=TFS_WriteFile(Text , sizeof(Text), fhandle);
                    if(written != sizeof(Text))
                        returnOk=TFS_GetErrNo();
                    if(returnOk==TFS_NO_ERR) {
                        returnOk=TFS_SeekFile(45, fhandle);
                        if(returnOk==TFS_NO_ERR) {
                            written=TFS_ReadFile((U08 OS_HUGE *)Readed, sizeof(Text), fhandle);
                            if(written != sizeof(Text))
                                returnOk=TFS_GetErrNo();
                        }
                    }
                }
                returnOk=TFS_CloseFile(fhandle);
                BSP_toggle_LED();                                      // Toggle LED
            } else {
                returnOk=TFS_GetErrNo();
                if(returnOk == TFS_NO_ERR) 
                    while(1);
            }
            returnOk=TFS_CloseUser();
        }
        OS_TimeDly(10);                                     // test timer-wakeup
    }
}
