/***********************************************************
*                       pC/TFS 2.38b                       *
*                 a FREE - Tiny File System                *
*                                                          *
*           by Frank Goetze  -  www.embedded-os.de         *
************************************************************
* Permission to use, copy, modify, and distribute this     *
* software in source and binary forms and its              *
* documentation for any purpose and without fee is hereby  *
* granted, provided that the above authors notice appear   *
* in all copies and that both that authors notice and this *
* permission notice appear in supporting documentation.    *
*                                                          *
* THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND   *
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT    *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY    *
* AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.     *
* IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,  *
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR             *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,    *
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF     *
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER *
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN        *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING           *
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE   *
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY *
* OF SUCH DAMAGE.                                          *
************************************************************
*                        pC_TFS.h                          *
*                 FILESYSTEM DECLARATIONS                  *
***********************************************************/
#ifndef _pC_TFS_
 #define _pC_TFS_


#include "TFS_cfg.h"                  // bind config of pC/TFS

/*
************************************************************
*                      ERROR CODES
************************************************************
*/
#define TFS_NO_ERR       0           // no errors
#define TFS_USR_OVF      200         // no user free
#define TFS_DBL_USER     201         // double user
#define TFS_NO_USER      202         // not a valid user
#define TFS_SUB_USER     203         // a user is registrated with this dir as ROOT

#define TFS_NAME_EXIST   210         // name of entry exist in this DIR
#define TFS_NOT_EXIST    211         // DIR or FILE not exist
#define TFS_PATH_ERR     212         // error on PATH
#define TFS_TMP_DIR      213         // Temp-Dir is set / not set
#define TFS_NO_FILE      214         // no FILE-Name given
#define TFS_FILE_RO      215         // file to open for writing is read-only
#define TFS_FILE_WO      216         // file to open for reading is write-only
#define TFS_FILE_EOF     217         // end-of-file while reading or writing
#define TFS_NOT_EMPTY    218         // entry not empty
#define TFS_FILE_OPEN    219         // current entry is opened file
#define TFS_NO_DATA      220         // current file-lenght is zero
#define TFS_WRONG_PTR    221         // offset into open file is wrong (or size for R/W)

#define TFS_LINKED       230         // entry is linked
#define TFS_MAX_LINK     231         // entry is max linked (255)
#define TFS_LINK_ERR     232         // error in link-mechanism
#define TFS_NO_LINK      233         // no link to an entry in link-entry

#define TFS_MEM_ERR      250         // error in memory-handle
#define TFS_MEM_OVF      251         // not enough memory
#define TFS_WR_PTR       252         // source/dest for write/read is in TFS-area (for EE-PROM)
#define TFS_WRITE_ERR    253         // memory write error

/*
************************************************************
*                ENTRY-TYPES / FILE ATTRIBUTES
************************************************************
*/
#define TFS_ATTR_DIR     0x80        // Entry is a directory
#define TFS_ATTR_FILE    0x40        // Entry is a file
#define TFS_ATTR_DB      0x20        // Entry is a data-block
#define TFS_ATTR_LINK    0x10        // Entry is a Linked element
#define TFS_ATTR_TYPE_M  0xF0        // Entry type mask

#define TFS_ATTR_RW      0x00        // Entry is read/write
#define TFS_ATTR_RO      0x01        // Entry is read-only
#define TFS_ATTR_WO      0x02        // Entry is write-only
#define TFS_ATTR_SYS     0x04        // Entry is system-state


/*
************************************************************
*                      FILE ACCESS-MODES
************************************************************
*/
#define TFS_ACC_RW       0x00        // File will read and write
#define TFS_ACC_RO       0x01        // File will read only
#define TFS_ACC_WO       0x02        // File will written only

#define TFS_SEEK_EOF    (TFS_LONG)(0-1)  // goto EndOfFile

#ifdef OS_PACK_PRAGMA
 #pragma pack(1)
#endif
/*
***********************************************************
*                   TFS MEMORY-STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_m {
#if (TFSDBLLINKED==1)
    struct tfs_m OS_HUGE *Prev;
#endif
    struct tfs_m OS_HUGE *Next;
    TFS_LONG              Len;
} OS_PACKED_ATTR TFS_MEM;

/*
***********************************************************
*                  TFS ENTRY SUB-STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_e {
    TFS_ATTR              Attr;
#if (TFSDBLLINKED==1)
    struct tfs_e OS_HUGE *Prev;
#endif
    struct tfs_e OS_HUGE *Next;
    struct tfs_e OS_HUGE *PrevLevel;
} OS_PACKED_ATTR TFS_ENTRY;

/*
***********************************************************
*                    TFS DIR/FILE STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_df {
    struct tfs_e          Entry;
    U08                   LinkCnt;
    U08                   Name[TFS_Name_SIZE];
    TFS_LONG              UsedLength;
    struct tfs_e OS_HUGE *Sub;
} OS_PACKED_ATTR TFS_DIRFILE;

/*
***********************************************************
*                  TFS DATA-ELEMENT STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_d {
    struct tfs_e          Entry;
    TFS_LONG              Length;
} OS_PACKED_ATTR TFS_DATA;

/*
***********************************************************
*                    OPENED FILE STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_o {
    struct tfs_e OS_HUGE *openEntry;
    U08                   openMode;
    TFS_LONG              openOffset;
} OS_PACKED_ATTR TFS_OPEN;

/*
***********************************************************
*                    TFS USER STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_user {
    struct tfs_e OS_HUGE *currentEntry;
#if (!(TFSDBLLINKED==1))
    struct tfs_e OS_HUGE *prevEntry;
#endif
    struct tfs_e OS_HUGE *getEntry;
#if (TFS_SubROOT==1)
    struct tfs_e OS_HUGE *SubROOT;
#endif
#if (TFS_TempDIR==1)
    struct tfs_e OS_HUGE *TempDir;
#endif
    struct tfs_o          used[TFS_HANDLES];
    U08                   ErrNo;
    U08                   Id;
} OS_PACKED_ATTR TFS_USER;

/*
***********************************************************
*                    TFS GET STRUCTURE
***********************************************************
*/
OS_PACKED typedef struct tfs_get {
    U08                   Name[TFS_Name_SIZE];
    TFS_ATTR              Attr;
    TFS_LONG              Length;
    U08                   Linked;
} OS_PACKED_ATTR TFS_GET;

#ifdef OS_PACK_PRAGMA
 #pragma pack()
#endif

/*
***********************************************************
*                   TFS FUNCTION PROTOTYPES
***********************************************************
*/
U08          TFS_Init(void) KC51_REENTRANT;
TFS_PATHNAME OS_HUGE *TFS_GetRev(void) KC51_REENTRANT;
U08          TFS_Format(void) KC51_REENTRANT;
TFS_LONG     TFS_GetFreeSize(void) KC51_REENTRANT;

U08          TFS_BecomeUser(TFS_USER OS_HUGE *TFSUser) KC51_REENTRANT;
U08          TFS_BecomeUserSubROOT(TFS_USER OS_HUGE *TFSUser, TFS_PATHNAME OS_HUGE *FullDirPath) KC51_REENTRANT;
U08          TFS_CloseUser(void) KC51_REENTRANT;

U08          TFS_CreateDir(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_RemoveDir(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_RenameDir(TFS_PATHNAME OS_HUGE *OldName, TFS_PATHNAME OS_HUGE *NewName) KC51_REENTRANT;
U08          TFS_ChangeDir(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_ChangeDirTemp(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_BackDirTemp(void) KC51_REENTRANT;
//          "TFS_MoveDir" not implemented, since links can show to sub-entries of this dir and so this links will moved too !

U08          TFS_CreateFile(TFS_PATHNAME OS_HUGE *Name, TFS_ATTR Attr, TFS_LONG size) KC51_REENTRANT;
U08          TFS_RemoveFile(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_AttribFile(TFS_PATHNAME OS_HUGE *Name, TFS_ATTR Attribs) KC51_REENTRANT;
U08          TFS_ResizeFile(TFS_PATHNAME OS_HUGE *Name, TFS_LONG size) KC51_REENTRANT;
U08          TFS_RenameFile(TFS_PATHNAME OS_HUGE *OldName, TFS_PATHNAME OS_HUGE *NewName) KC51_REENTRANT;
U08          TFS_MoveFile(TFS_PATHNAME OS_HUGE *Name, TFS_PATHNAME OS_HUGE *NewPath) KC51_REENTRANT;
TFS_ATTR     TFS_GetFileAttrib(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
TFS_LONG     TFS_GetFileSize(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;

TFS_HANDLE   TFS_OpenFile(TFS_PATHNAME OS_HUGE *Name, U08 Mode) KC51_REENTRANT;
U08          TFS_SeekFile(TFS_LONG offset, TFS_HANDLE handl) KC51_REENTRANT;
U08          TFS_ExpandFile(TFS_LONG addsize, TFS_HANDLE handl) KC51_REENTRANT;
U08          TFS_SetEOF(TFS_HANDLE handl) KC51_REENTRANT;
TFS_LONG     TFS_TellFile(TFS_HANDLE handl) KC51_REENTRANT;
TFS_LONG     TFS_ReadFile(U08 OS_HUGE *dest, TFS_LONG size, TFS_HANDLE handl) KC51_REENTRANT;
TFS_LONG     TFS_WriteFile(U08 OS_HUGE *src, TFS_LONG size, TFS_HANDLE handl) KC51_REENTRANT;
TFS_LONG     TFS_WriteFileE(U08 OS_HUGE *src, TFS_LONG size, TFS_HANDLE handl) KC51_REENTRANT;
U08          TFS_GetErrNo(void) KC51_REENTRANT;
U08          TFS_CloseFile(TFS_HANDLE handl) KC51_REENTRANT;

U08          TFS_GetCurrentDir(U08 OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_GetCurrentPath(U08 OS_HUGE *Path, TFS_LONG maxlen) KC51_REENTRANT;
U08          TFS_GetFirstName(U08 OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_GetNextName(U08 OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_GetFirst(TFS_GET OS_HUGE *get) KC51_REENTRANT;
U08          TFS_GetNext(TFS_GET OS_HUGE *get) KC51_REENTRANT;
U08          TFS_GetEntry(TFS_PATHNAME OS_HUGE *Name, TFS_GET OS_HUGE *get) KC51_REENTRANT;

U08          TFS_CreateLink(TFS_PATHNAME OS_HUGE *OrgName, TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_RemoveLink(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT;
U08          TFS_ReadLink(TFS_PATHNAME OS_HUGE *Name, U08 OS_HUGE *LinkName) KC51_REENTRANT;

#if defined(_WIN32) || defined(WIN32) || defined(__LINUX__) || defined(__linux__) || defined(__CYGWIN__)
 S16         TFS_Flush(void);        // external function from Window/LINUX extender
#endif

/************************* END ****************************/
#endif




