/***********************************************************
*                       pC/TFS 2.38b                       *
*                 a FREE - Tiny File System                *
*                                                          *
*            by Frank Goetze  -  www.embedded-os.de        *
************************************************************
*                    TFS_Port_Atmel_EE.c                   *
*                     EEPROM Atmel Port                    *
************************************************************
* Revision-Info:                                           *
*                                                          *
***********************************************************/

#include "../../inc/AT28/TFS_Port_Atmel_EE.h"

/*
************************************************************
*                    INIT MEMORY
************************************************************
*/
U08 TFSPort_Init(void)
{
    return(TFS_NO_ERR);                                    // nothing to init here
}

/*
************************************************************
*                    WRITE INTO EE
************************************************************
*/
U08 TFSPort_Write(U08 OS_HUGE *dest, OS_CONST U08 OS_HUGE *source, TFS_LONG length) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    TFS_LONG     byte;

    if((source >= (U08 OS_HUGE *)TFS_START) && (source < (U08 OS_HUGE *)(TFS_START + TFS_SIZE)))
        return(TFS_WR_PTR);                                // ERROR - source is in TFS-area
    if((dest < (U08 OS_HUGE *)TFS_START) || ((dest + length) > (U08 OS_HUGE *)(TFS_START + TFS_SIZE)))
        return(TFS_WR_PTR);                                // ERROR - dest is not in TFS-area
    byte = 0;
    while(byte < length) {
        OS_ENTER_CRITICAL();
        ((U08 OS_HUGE *)TFS_START)[0x5555] = 0xAA;         // write-sequence
        ((U08 OS_HUGE *)TFS_START)[0x2AAA] = 0x55;
        ((U08 OS_HUGE *)TFS_START)[0x5555] = 0xA0;
        do {
            dest[byte++] = source[byte];
        } while((((U32)(dest) - TFS_SIZE + byte) & (TFS_PAGESIZE - 1) != 0) && (byte < length));
        while((dest[byte - 1] & 0x80) != (source[byte - 1] & 0x80));
        OS_EXIT_CRITICAL();
    }                                                      // up to page-size and read flag
    return(TFS_NO_ERR);
}

/*
************************************************************
*                         FILL EE
************************************************************
*/
U08 TFSPort_Fill(U08 OS_HUGE *dest, U08 value, TFS_LONG length) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    TFS_LONG     byte;

    if((dest < (U08 OS_HUGE *)TFS_START) || ((dest + length) > (U08 OS_HUGE *)(TFS_START + TFS_SIZE)))
        return(TFS_WR_PTR);                                // ERROR - dest is not in TFS-area
    byte = 0;
    while(byte < length) {
        OS_ENTER_CRITICAL();
        ((U08 OS_HUGE *)TFS_START)[0x5555] = 0xAA;         // write-sequence
        ((U08 OS_HUGE *)TFS_START)[0x2AAA] = 0x55;
        ((U08 OS_HUGE *)TFS_START)[0x5555] = 0xA0;
        do {
            dest[byte++] = value;
        } while((((U32)(dest) - TFS_SIZE + byte) & (TFS_PAGESIZE - 1)) != 0 && (byte < length));
        while((dest[byte - 1] & 0x80) != (value & 0x80));
        OS_EXIT_CRITICAL();
    }                                                      // up to page-size and read flag
    return(TFS_NO_ERR);
}
