/***********************************************************
*                       pC/TFS 2.38b                       *
*                 a FREE - Tiny File System                *
*                                                          *
*           by Frank Goetze  -  www.embedded-os.de         *
************************************************************
* Permission to use, copy, modify, and distribute this     *
* software in source and binary forms and its              *
* documentation for any purpose and without fee is hereby  *
* granted, provided that the above authors notice appear   *
* in all copies and that both that authors notice and this *
* permission notice appear in supporting documentation.    *
*                                                          *
* THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND   *
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT    *
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY    *
* AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.     *
* IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT,  *
* INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR             *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,    *
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF     *
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER *
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN        *
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING           *
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE   *
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY *
* OF SUCH DAMAGE.                                          *
************************************************************
*                        pC_TFS.c                          *
*                       File System                        *
************************************************************
* Revision-Info:                                           *
*                                                          *
* 1.00a     first alpa-release                             *
* 1.20d     some little bug-fixes in TFS_GetFirstEntry and *
*           TFS_GetNextEntry and two additional file-      *
*           attributes. Directories havn't attribs.        *
*           Optimising code.                               *
* 2.00a     alpha-release with links. Tripple-source for   *
*           Keil, gcc and Microsoft-C                      *
* 2.01b     Changed return-parameter of TFS_OpenFile,      *
*           TFS_FileRead and TFS_FileWrite. Implementation *
*           of TFS_GetErrNo(). Implementation of a         *
*           additional LINUX-Host port's to simulate the   *
*           TFS in RAM of a PC.                            *
* 2.10a     Userhandling changed - every Task can only be  *
*           one User, but every User can open TFS_HANDLES  *
*           files. Adaption of API-Calls.                  *
*           Function of TFS_ExpandFile() changed, new      *
*           function TFS_ResizeFile() added.               *
*           For a FileSystem smaller than 64kByte TFS_LONG *
*           for all length-parameter implemented to        *
*           minimize the API- and Memory- Overhead.        *
*           Now a file can opened simultaneously from many *
*           Users to reading - but only one User can open  *
*           a file to writing.                             *
*           If a file opened (more than once) to reading,  *
*           an other User can't open this file to writing  *
*           - if a file from one User opened to writing,   *
*           no other User can open this file for reading   *
*           or writing.                                    *
* 2.11b     TFS_GetFileSize, TFS_GetFileAttrib and         *
*           TFS_TellFile added. Little bugfix in internal  *
*           function TFSChange_Entry.                      *
* 2.20b     complete code optimizing and clean-up          *
*           tests with MSC-8.00 / GCC / Keil-C51 /         *
*           ImageCraft / Tasking / IAR / ADS / Keil-RDS    *
* 2.21a     TFS_GetFreeSize and TFS_MoveFile added         *
*           bugfix in TFS_RenameFile and TFS_RenameDir     *
* 2.23b     TFS_WriteFileE added. SpeedUp in internal      *
*           function TFSChange_Entry.                      *
* 2.30d     complete update to U08..U32 types.             *
*           changes of HANDLE to TFS_HANDLE , TFSLONG to   *
*           TFS_LONG and ATTR to TFS_ATTR.                 *
*           small correction in TFS_SeekFile().            *
*           better portability for OS-less use implemented *
*           clean-up in comments and some lines            *
* 2.32a     TFS_GetFirstEntry and TFS_GetNextEntry renamed *
*           to TFS_GetFirstName and TFS_GetNextName.       *
*           TFS_GetFirst and TFS_GetNext added, this       *
*           functions writes all infos about an entry      *
*           (DIR/FILE) in a given structure.               *
*           TFS_GetRev added                               *
* 2.34b     TFS_BecomeUserSubROOT, TFS_GetCurrentPath,     *
*           TFS_ChangeDirTemp and TFS_BackDirTemp added    *
* 2.35a     change to task-ID on use of pC/OS for unique   *
*           identify of running task, TFS_SetEOF added     *
* 2.36a     correct user-check for GCC & unknown user      *
* 2.37b     TFS_GetEntry added and correction in internal  *
*           function TFSChange_Entry                       *
* 2.38b     TFS_PATHNAME typedef replaces all string       *
*           inputs to centralize them for compiler         *
*           sensitivities                                  *
*                                                          *
***********************************************************/

/*
************************************************************
*                    additional bindings                   *
*             (TFS itself is binded by the port)           *
************************************************************
*/
#include "../inc/TFS_cfg.h"
#include "../inc/TFS_string.h"

#ifdef _pC_OS_
 extern OS_TCB* OSTCBCur;
#endif

#ifndef TFS_START
 extern U08 TFS_START[TFS_SIZE];
#endif

/***************** Constants im Code **********************/

TFS_PATHNAME OS_ROMDATA TFSRevision[] = {"pC/TFS 2.38b"};

TFS_DIRFILE OS_CONST OS_ROMDATA ROOT = {                   // ------- ROOT-Entry -map --------
        { 0x80,                                            // Attribs (Entry is a directory)
#if (TFSDBLLINKED==1)
        (TFS_ENTRY OS_HUGE *)NULL,                         // prev-PTR
#endif
        (TFS_ENTRY OS_HUGE *)NULL,                         // next-PTR
        (TFS_ENTRY OS_HUGE *)NULL },                       // prevLevel-PTR
        0,                                                 // LinkCnt
        { '/',0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 },             // Name ("TFS_Name_SIZE" dependend)
        0,                                                 // Used-Lenght (not used for DIRs)
        (TFS_ENTRY OS_HUGE *)NULL                          // Sub-PTR
};

/****************** Globals im RAM ************************/

#if defined(_WIN32) || defined(WIN32) || defined(__LINUX__) || defined(__linux__) || defined(__CYGWIN__)
        TFS_DIRFILE OS_HUGE *ROOT_PTR;                     // begin of TFS
#else
 static TFS_DIRFILE OS_HUGE *ROOT_PTR;                     // begin of TFS
#endif
static TFS_USER    OS_HUGE *TFS_User[TFS_MAX_USER];        // user control
static TFS_MUTEX_TYPE       TFS_mutex;                     // the TFS-mutex


/**************** Forward Prototypes **********************/

static U08      TFSCheckUser(U08 *id) KC51_REENTRANT;
static U08      TFSFileOpend(TFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT;
static U08      TFSChange_Entry(TFS_PATHNAME OS_HUGE *Name, U08 mode) KC51_REENTRANT;
static U08      TFSCreate_Entry(TFS_PATHNAME OS_HUGE *Name, TFS_ATTR Attribs, TFS_LONG size) KC51_REENTRANT;
#if (TFSDBLLINKED==1)
 static U08     TFSUnlink_Entry(TFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT;
 static U08     TFSDelete_Entry(TFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT;
#else
 static U08     TFSUnlink_Entry(TFS_ENTRY OS_HUGE *EntryPTR, TFS_ENTRY OS_HUGE *PrevPTR) KC51_REENTRANT;
 static U08     TFSDelete_Entry(TFS_ENTRY OS_HUGE *EntryPTR, TFS_ENTRY OS_HUGE *PrevPTR) KC51_REENTRANT;
#endif
static U08      TFSResize_File(TFS_DIRFILE OS_HUGE *FilePTR, TFS_LONG size) KC51_REENTRANT;
static U08      TFSExpand_File(TFS_DIRFILE OS_HUGE *FilePTR, TFS_LONG size) KC51_REENTRANT;

static U08      TFSMem_Format(void) KC51_REENTRANT;
static TFS_LONG TFSMem_FreeSize(void) KC51_REENTRANT;
static U08      TFSMem_Alloc(TFS_ENTRY OS_HUGE *(*TFSMPtr), TFS_LONG size, TFS_LONG *biggest) KC51_REENTRANT;
static U08      TFSMem_Add(TFS_MEM OS_HUGE *TFSMPtr, TFS_LONG size) KC51_REENTRANT;
static U08      TFSMem_Free(TFS_MEM OS_HUGE *TFSMPtr) KC51_REENTRANT;

/*
************************************************************
*              INITIALIZE TFS-MUTEX (internal)             *
************************************************************
*/
static U08 TFSMutex_Init(TFS_MUTEX_TYPE OS_HUGE *mutex) KC51_REENTRANT
{
#ifdef _pC_OS_
    return(OS_SemInit(mutex, 1));                          // use a semaphore on pC/OS to test sheduling during two task in TFS
#else
    *mutex = 0; return(TFS_NO_ERR);
#endif
}
/*
************************************************************
*                 GET TFS-MUTEX (internal)                 *
************************************************************
*/
static U08 TFSMutex_Get(TFS_MUTEX_TYPE OS_HUGE *mutex) KC51_REENTRANT
{
#ifdef _pC_OS_
    return(OS_SemPend(mutex, OS_SUSPEND));                 // use a semaphore on pC/OS to test sheduling during two task in TFS
#else
    if (*mutex) return(1);                                 // as mark of "in use"
    //@TODO: disable interrupts here !
    *mutex = 1;
    //@TODO: enable interrupts here !
    return(TFS_NO_ERR);
#endif
}
/*
************************************************************
*                 PUT TFS-MUTEX (internal)                 *
************************************************************
*/
static U08 TFSMutex_Put(TFS_MUTEX_TYPE OS_HUGE *mutex) KC51_REENTRANT
{
#ifdef _pC_OS_
    return(OS_SemPost(mutex));
#else
    *mutex = 0; return(TFS_NO_ERR);
#endif
}

/*
************************************************************
*                    INITIALIZE TFS                        *
************************************************************
*/
U08 TFS_Init(void) KC51_REENTRANT
{
    U08 returnOk, User;

    returnOk = TFSMutex_Init(&TFS_mutex);
    if(returnOk == TFS_NO_ERR) {
        returnOk = TFSPort_Init();
        if(returnOk == TFS_NO_ERR) {
            ROOT_PTR = (TFS_DIRFILE OS_HUGE *)((U08 OS_HUGE *)(TFS_START) + sizeof(TFS_MEM));
            for(User = 0; User < TFS_MAX_USER; User++) {
                TFS_User[User] = (TFS_USER OS_HUGE *)NULL;
            }
#if defined(_WIN32) || defined(WIN32) || defined(__LINUX__) || defined(__linux__) || defined(__CYGWIN__)
            returnOk = TFS_Load();                         // load DRIVE-Image from a real file into RAM-DRIVE (only for RAM-Port!)
#endif
            if((TFSstrcmp((TFS_PATHNAME OS_HUGE *)ROOT_PTR->Name, (TFS_PATHNAME OS_HUGE *)(ROOT.Name)))
#if (TFSDBLLINKED==1)
              || (ROOT_PTR->Entry.Prev != (TFS_ENTRY OS_HUGE *)NULL)
#endif
              || (ROOT_PTR->Entry.Next != (TFS_ENTRY OS_HUGE *)NULL)
              || (ROOT_PTR->Entry.PrevLevel != (TFS_ENTRY OS_HUGE *)NULL)) {
                returnOk = TFS_Format();                   // set ROOT, check drive and format..
            }
        }
    }
    return(returnOk);                                      // ..is not valid
}

/*
************************************************************
*                    GET REVISION                          *
************************************************************
*/
TFS_PATHNAME OS_HUGE *TFS_GetRev(void) KC51_REENTRANT
{
    return(TFSRevision);
}

/*
************************************************************
*                       FORMAT DRIVE                       *
************************************************************
*/
U08 TFS_Format(void) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE *EntryPTR;
    U08                returnOk, i, j;
    TFS_LONG           big_null;

    returnOk = TFSMutex_Get(&TFS_mutex);
    if(returnOk == TFS_NO_ERR) {
        for(i = 0; i < TFS_MAX_USER; i++) {
            for(j = 0; j < TFS_HANDLES; j++) {
                if((TFS_User[i] != (TFS_USER OS_HUGE *)NULL) &&
                   (TFS_User[i]->used[j].openEntry != (TFS_ENTRY OS_HUGE *)NULL)) {
                    return(TFS_FILE_OPEN);
                }
            }
        }
        returnOk = TFSMem_Format();                        // format the MEM-level
        if(returnOk == TFS_NO_ERR) {                       // create ROOT-entry
            returnOk = TFSMem_Alloc(&EntryPTR, sizeof(TFS_DIRFILE), &big_null);
            if(returnOk == TFS_NO_ERR) {
                returnOk = TFSPort_Write((U08 OS_HUGE *)(EntryPTR), (U08 OS_HUGE *)(&ROOT), sizeof(TFS_DIRFILE));
            }
            if(returnOk == TFS_NO_ERR) {
                for(i = 0; i < TFS_MAX_USER; i++) {
                    if(TFS_User[i] != (TFS_USER OS_HUGE *)NULL) {
                        TFS_User[i]->currentEntry = (TFS_ENTRY OS_HUGE *)(ROOT_PTR);
                        TFS_User[i]->getEntry     = (TFS_ENTRY OS_HUGE *)NULL;
                        TFS_User[i]->ErrNo        = TFS_NO_ERR;
                        for(j = 0; j < TFS_HANDLES; j++) {
                            TFS_User[i]->used[j].openEntry = (TFS_ENTRY OS_HUGE *)NULL;
                            TFS_User[i]->used[j].openMode  = 0;
                            TFS_User[i]->used[j].openOffset= 0;
                        }
                    }
                }
            }
        }
        returnOk = TFSMutex_Put(&TFS_mutex);
    }
    return(returnOk);
}

/*
************************************************************
*                GET USER BY ID (internal)                 *
************************************************************
*/
static U08 TFSGetUserByID(void) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08 i, id;

    OS_ENTER_CRITICAL();
    id = OS_USER_ID;
    for(i = 0; i < TFS_MAX_USER; i++) {
        if((TFS_User[i] != (TFS_USER OS_HUGE *)NULL) && (TFS_User[i]->Id == id)) {
            OS_EXIT_CRITICAL();
            return(i);
        }
    }
    OS_EXIT_CRITICAL();
    return(0xFF);
}

/*
************************************************************
*                 GET FREE USER (internal)                 *
************************************************************
*/
static U08 TFSGetFreeUser(U08 *user) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08 i;

    OS_ENTER_CRITICAL();
    i = TFSGetUserByID();
    if(i == 0xFF) {
        for(i = 0; i < TFS_MAX_USER; i++) {
            if(TFS_User[i] == (TFS_USER OS_HUGE *)NULL) {
                break;
            }
        }
        if(i < TFS_MAX_USER) {
            *user = i;
            OS_EXIT_CRITICAL();
            return(TFS_NO_ERR);
        }
        i = 0xFF;
    }
    if(i == 0xFF) {
        OS_EXIT_CRITICAL();
        return(TFS_USR_OVF);
    }
    OS_EXIT_CRITICAL();
    return(TFS_DBL_USER);
}

/*
************************************************************
*                 CHECK USER (internal)                    *
************************************************************
*/
static U08 TFSCheckUser(U08 *user) KC51_REENTRANT
{
    *user = TFSGetUserByID();
    if(*user != 0xFF) {
        if(TFS_User[*user] != (TFS_USER OS_HUGE *)NULL) {
            TFS_User[*user]->ErrNo = TFSMutex_Get(&TFS_mutex);
            return(TFS_User[*user]->ErrNo);
        } else {
            *user = 0xFF;
        }
    }
    return(TFS_NO_USER);
}

/*
************************************************************
*               GET FREE-SIZE OF MEMORY (BRUTTO)           *
************************************************************
*/
TFS_LONG TFS_GetFreeSize(void) KC51_REENTRANT
{
    TFS_LONG i;
    U08      user;

    if(TFS_NO_ERR != TFSCheckUser(&user)) {
        return(0);
    }
    i = TFSMem_FreeSize();
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    if(TFS_User[user]->ErrNo != TFS_NO_ERR) {
        return(0);
    }
    return(i);
}

/*
************************************************************
*                        BECOME USER                       *
************************************************************
*/
U08 TFS_BecomeUser(TFS_USER OS_HUGE *TFSUser) KC51_REENTRANT
{
#if OS_CRITICAL_SR
    OS_STK_TYPE  OS_sr;
#endif
    U08          i, user;

    OS_ENTER_CRITICAL();
    i = TFSGetFreeUser(&user);
    if(i != TFS_NO_ERR) {
        OS_EXIT_CRITICAL();
        return(i);
    }
    if(TFS_User[user] == (TFS_USER OS_HUGE *)NULL) {
        TFS_User[user]               = TFSUser;
        TFS_User[user]->Id           = OS_USER_ID;
        OS_EXIT_CRITICAL();
        TFS_User[user]->currentEntry = (TFS_ENTRY OS_HUGE *)ROOT_PTR;
        TFS_User[user]->getEntry     = (TFS_ENTRY OS_HUGE *)NULL;
#if (TFS_SubROOT==1)
        TFS_User[user]->SubROOT      = (TFS_ENTRY OS_HUGE *)NULL;
#endif
#if (TFS_TempDIR==1)
        TFS_User[user]->TempDir      = (TFS_ENTRY OS_HUGE *)NULL;
#endif
        TFS_User[user]->ErrNo        = TFS_NO_ERR;
        for(i = 0; i < TFS_HANDLES; i++) {
            TFS_User[user]->used[i].openEntry = (TFS_ENTRY OS_HUGE *)NULL;
            TFS_User[user]->used[i].openMode  = 0;
            TFS_User[user]->used[i].openOffset= 0;
        }
        return(TFS_NO_ERR);
    } else {
        OS_EXIT_CRITICAL();
        return(TFS_DBL_USER);
    }
}

#if (TFS_SubROOT==1)
/*
************************************************************
*               BECOME USER WITH SUBDIR AS ROOT            *
************************************************************
*/
U08 TFS_BecomeUserSubROOT(TFS_USER OS_HUGE *TFSUser, TFS_PATHNAME OS_HUGE *FullDirPath) KC51_REENTRANT
{
    U08 returnOk, user;

    if(FullDirPath[0] != '/') {
        return(TFS_PATH_ERR);
    }
    returnOk = TFS_BecomeUser(TFSUser);                    // create user
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSMutex_Get(&TFS_mutex);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSChange_Entry(FullDirPath, 5);            // go into dir / into linked dir
    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);                               // error ! - cant find the user
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    if(returnOk != TFS_NO_ERR) {
        TFS_User[user] = (TFS_USER OS_HUGE *)NULL;         // clear user
        return(returnOk);
    }
    TFS_User[user]->SubROOT = TFS_User[user]->currentEntry;// set this DIR as ROOT
    return(TFS_User[user]->ErrNo);
}
#endif

/*
************************************************************
*                        CLOSE USER                        *
************************************************************
*/
U08 TFS_CloseUser(void) KC51_REENTRANT
{
    U08 i, user;

    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);
    }
    for(i = 0; i < TFS_HANDLES; i++) {
        if(TFS_User[user]->used[i].openEntry != (TFS_ENTRY OS_HUGE *)NULL) {
            return(TFS_FILE_OPEN);
        }
    }
    TFS_User[user] = (TFS_USER OS_HUGE *)NULL;
    return(TFS_NO_ERR);
}

/*
************************************************************
*     GIVE THE ERROR-CODE FROM OPEN / READ / WRITE / ..    *
************************************************************
*/
U08 TFS_GetErrNo(void) KC51_REENTRANT
{
    U08 returnOk, user;

    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);
    }
    returnOk = TFS_User[user]->ErrNo;
    TFS_User[user]->ErrNo = TFS_NO_ERR;
    return(returnOk);
}

/*
************************************************************
*                         OPEN FILE                        *
************************************************************
*/
TFS_HANDLE TFS_OpenFile(TFS_PATHNAME OS_HUGE *Name, U08 Mode) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE   *orgPTR;
    U08                  returnOk, i, j, user;

    if(TFS_NO_ERR != TFSCheckUser(&user)) {
        return(0);                                         // no correct user
    }
    j = (U08)-1;
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position
    returnOk = TFSChange_Entry(Name, 6);                   // go into file / into linked file
    if(returnOk == TFS_NO_ERR) {                           // go into PATH
        i = ((U08)(TFS_User[user]->currentEntry->Attr)) & (TFS_ATTR_RO | TFS_ATTR_WO);
        if(i == Mode || i == 0) {
            for(i = 0; i < TFS_MAX_USER; i++) {
                if(TFS_User[i] != (TFS_USER OS_HUGE *)NULL) {
                    for(j = 0; j < TFS_HANDLES; j++) {
                        if((TFS_User[i]->used[j].openEntry == TFS_User[user]->currentEntry) &&
                          ((TFS_User[i]->used[j].openMode != TFS_ACC_RO) ||
                           (Mode != TFS_ACC_RO))) {
                            TFS_User[user]->currentEntry = orgPTR;  // restore actual position
                            TFS_User[user]->ErrNo = TFS_FILE_OPEN;
                            TFSMutex_Put(&TFS_mutex);
                            return(0);
                        }
                    }
                }
            }
            if(returnOk == TFS_NO_ERR) {
                i = 0; j = TFS_HANDLES;
                while(i < j) {
                    if((TFS_User[user]->used[i].openEntry == (TFS_ENTRY OS_HUGE *)NULL) &&
                       (j == TFS_HANDLES)) {
                        j = i;
                    }
                    i++;
                }
                if(j != TFS_HANDLES) {
                    TFS_User[user]->used[j].openEntry  = TFS_User[user]->currentEntry;
                    TFS_User[user]->used[j].openMode   = Mode;
                    TFS_User[user]->used[j].openOffset = 0;
                } else {                                   // no free handle-field
                    returnOk = TFS_USR_OVF;
                }
            }
        } else {                                           // file is RO or WO
            if(Mode == TFS_ACC_RO) {
                returnOk = TFS_FILE_WO;
            } else {
                returnOk = TFS_FILE_RO;
            }
        }
    }
    TFS_User[user]->currentEntry = orgPTR;                 // restore actual position
    TFS_User[user]->ErrNo = returnOk;
    TFSMutex_Put(&TFS_mutex);
    if(returnOk == TFS_NO_ERR) {
        return((TFS_HANDLE)(j+1));
    }
    return(0);
}

/*
************************************************************
*                         CLOSE FILE                       *
************************************************************
*/
U08 TFS_CloseFile(TFS_HANDLE handl) KC51_REENTRANT
{
    U08 returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > TFS_HANDLES) ||
       (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(TFS_NO_USER);                               // if actual position opened file - go one back
    }
    handl -= 1;
    if((TFS_User[user]->used[handl].openEntry->Attr) & TFS_ATTR_FILE) {
        TFS_User[user]->used[handl].openEntry  = (TFS_ENTRY OS_HUGE *)NULL;
        TFS_User[user]->used[handl].openMode   = 0;
        TFS_User[user]->used[handl].openOffset = 0;
        returnOk = TFS_NO_ERR;
    } else {
        returnOk = TFS_NO_FILE;                            // actual position is no file
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         RENAME FILE                      *
************************************************************
*/
U08 TFS_RenameFile(TFS_PATHNAME OS_HUGE *OldName, TFS_PATHNAME OS_HUGE *NewName) KC51_REENTRANT
{
    TFS_ENTRY   OS_HUGE *orgPTR;
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_DIRFILE OS_HUGE *DirPTR;
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  returnOk, user;
    TFS_LONG             Pos, i;

    Pos = 0;
    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(OldName, 3);                // go into file / stay on link-file
    if(returnOk == TFS_NO_ERR) {                           // if valid old name - go into
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
        if(returnOk == TFS_NO_ERR) {
            FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
            DirPTR  = (TFS_DIRFILE OS_HUGE *)(FilePTR->Entry.PrevLevel);
            Pos = TFSstrrpos(NewName,'/');                 // search last '/' in PATH or full name and check
            if(Pos == (TFS_LONG)(0-1)) {
                Pos = 0;
            } else {
                Pos++;                                     // update name if NewName not exist
            }
            if((Pos < TFSstrlen(NewName)) && (TFSstrlen(NewName)-Pos < TFS_Name_SIZE)) {
                EntryPTR = (TFS_DIRFILE OS_HUGE *)(DirPTR->Sub);   // get first entry and check if name exist
                while((0 != (i = TFSstrcmp(&NewName[Pos], (TFS_PATHNAME OS_HUGE *)EntryPTR->Name))) && (NULL != EntryPTR->Entry.Next)) {
                    EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
                }
                if(i == 0) {                               // up to end of level
                    returnOk = TFS_NAME_EXIST;             // NewName exist
                } else {
                    returnOk = TFSPort_Write((U08 OS_HUGE *)FilePTR->Name, (U08 OS_HUGE *)&NewName[Pos], TFSstrlen(&NewName[Pos])+1);
                }
            } else {
                if(Pos == TFSstrlen(NewName)) {
                    returnOk = TFS_NO_FILE;                // no new-filename given
                } else {
                    returnOk = TFS_PATH_ERR;               // file not exist
                }
            }
        }
    }
    TFS_User[user]->currentEntry = orgPTR;                 // restore actual position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                    CHANGE FILE-ATTRIBS                   *
************************************************************
*/
U08 TFS_AttribFile(TFS_PATHNAME OS_HUGE *Name, TFS_ATTR Attribs) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE   *orgPTR;
    U08                  returnOk, user;
    TFS_ATTR             a;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 3);                   // go into file  / stay on link-file
    if(returnOk == TFS_NO_ERR) {                           // if valid path and filename - go into
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
        if(returnOk == TFS_NO_ERR) {
            a  = (TFS_User[user]->currentEntry->Attr & TFS_ATTR_LINK) | (Attribs & (TFS_ATTR_RO | TFS_ATTR_WO |TFS_ATTR_SYS)); 
            a |= TFS_ATTR_FILE;
            returnOk = TFSPort_Write((U08 OS_HUGE *)&(TFS_User[user]->currentEntry->Attr), (U08 OS_HUGE *)&a, sizeof(TFS_ATTR));
        }
    }
    TFS_User[user]->currentEntry = orgPTR;
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                      GET FILE-ATTRIBS                    *
************************************************************
*/
TFS_ATTR TFS_GetFileAttrib(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE   *orgPTR;
    TFS_ATTR             a;
    U08                  returnOk, user;

    if(TFS_NO_ERR != TFSCheckUser(&user)) {
        return(0);                                         // no correct user
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 3);                   // go into file  / stay on link-file
    if(returnOk == TFS_NO_ERR) {
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
        if(returnOk == TFS_NO_ERR) {
            a = TFS_User[user]->currentEntry->Attr & (TFS_ATTR_RO | TFS_ATTR_WO |TFS_ATTR_SYS);
            TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
            return(a);
        }
    }
    TFS_User[user]->ErrNo = returnOk;
    TFS_User[user]->currentEntry = orgPTR;
    TFSMutex_Put(&TFS_mutex);
    return((TFS_ATTR)(-1));
}

/*
************************************************************
*                      RESIZE FILE                         *
************************************************************
*/
U08 TFS_ResizeFile(TFS_PATHNAME OS_HUGE *Name, TFS_LONG size) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE   *orgPTR;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 6);                   // go into file / into linked file
    if(returnOk == TFS_NO_ERR) {
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
        if(returnOk == TFS_NO_ERR) {
            returnOk = TFSResize_File((TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry), size);
        }
    }
    TFS_User[user]->currentEntry = orgPTR;                 // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                   GET SIZE OF FILE                       *
************************************************************
*/
TFS_LONG TFS_GetFileSize(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_ENTRY   OS_HUGE *orgPTR;
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_LONG             len;
    U08                  returnOk, user;

    if(TFS_NO_ERR != TFSCheckUser(&user)) {
        return(0);
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 6);                   // go into file / into linked file
    if(returnOk == TFS_NO_ERR) {                           // if valid name - go into
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
    }
    FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
    TFS_User[user]->currentEntry = orgPTR;                 // restore saved position
    TFS_User[user]->ErrNo = returnOk;
    if(returnOk != TFS_NO_ERR) {
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // error
    }
    len = FilePTR->UsedLength;
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(len);
}

/*
************************************************************
*              MOVE FILE TO NEW POSITION IN TREE           *
************************************************************
*/
U08 TFS_MoveFile(TFS_PATHNAME OS_HUGE *Name, TFS_PATHNAME OS_HUGE *NewPath) KC51_REENTRANT
{
    TFS_ENTRY   OS_HUGE *orgPTR;
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_ENTRY   OS_HUGE *prePTR;
    TFS_DIRFILE OS_HUGE *DirPTR;
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  returnOk, user, i;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 3);                   // go into file / stay on link-file
    if(returnOk == TFS_NO_ERR) {                           // if valid name - go into
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
        FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
        if((returnOk == TFS_NO_ERR) && (FilePTR->LinkCnt == 0)) {
#if (!(TFSDBLLINKED==1))
            prePTR = TFS_User[user]->prevEntry;
#endif
            TFS_User[user]->currentEntry = orgPTR;         // restore saved position
            returnOk = TFSChange_Entry(NewPath, 5);        // go into dir  / into linked dir
            if(returnOk == TFS_NO_ERR) {                   // in currcent DIR
                DirPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
                if(DirPTR->Sub != NULL) {                  // get first entry and check if name exist
                    EntryPTR = (TFS_DIRFILE OS_HUGE *)(DirPTR->Sub);
                    while((0 != (i = TFSstrcmp((TFS_PATHNAME OS_HUGE *)FilePTR->Name, (TFS_PATHNAME OS_HUGE *)EntryPTR->Name))) && (NULL != EntryPTR->Entry.Next)) {
                        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
                    }
                    if(i == 0) {                           // up to end of level
                        TFS_User[user]->currentEntry = orgPTR;  // restore saved position and ABORT
                        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
                        return(TFS_NAME_EXIST);            // Entry-name exist
                    }
                }
#if (TFSDBLLINKED==1)                                      // unlink File in old position
                returnOk = TFSUnlink_Entry((TFS_ENTRY OS_HUGE *)(FilePTR));
#else
                returnOk = TFSUnlink_Entry((TFS_ENTRY OS_HUGE *)(FilePTR), prePTR);
#endif
                if(returnOk == TFS_NO_ERR) {               // update File-Entries
                    returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->Entry.PrevLevel), (U08 OS_HUGE *)(&DirPTR), sizeof(U08 OS_HUGE *));
                    if(returnOk == TFS_NO_ERR) {
                        prePTR = (TFS_ENTRY *)NULL;
                        returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->Entry.Next), (U08 OS_HUGE *)(&prePTR), sizeof(U08 OS_HUGE *));
#if (TFSDBLLINKED==1)
                        if(returnOk == TFS_NO_ERR) {
                            if(DirPTR->Sub == NULL) {
                                returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->Entry.Prev), (U08 OS_HUGE *)(&prePTR), sizeof(U08 OS_HUGE *));
                            } else {
                                returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->Entry.Prev), (U08 OS_HUGE *)(&EntryPTR), sizeof(U08 OS_HUGE *));
                            }
                        }
#endif
                    }
                }
                if(returnOk == TFS_NO_ERR) {               // link File in new position
                    if(DirPTR->Sub == NULL) {              // update hirarchical pointer-entries
                        returnOk = TFSPort_Write((U08 OS_HUGE *)(&DirPTR->Sub), (U08 OS_HUGE *)(&FilePTR), sizeof(U08 OS_HUGE *));
                    } else {
#if (TFSDBLLINKED==1)
                        returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->Entry.Prev->Next), (U08 OS_HUGE *)(&FilePTR), sizeof(U08 OS_HUGE *));
#else
                        returnOk = TFSPort_Write((U08 OS_HUGE *)(&EntryPTR->Entry.Next), (U08 OS_HUGE *)(&FilePTR), sizeof(U08 OS_HUGE *));
#endif
                    }
                }
            }
        } else {
            if(returnOk == TFS_NO_ERR) {
                returnOk = TFS_LINKED;
            }
        }
    }
    TFS_User[user]->currentEntry = orgPTR;                 // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                    EXPAND OPEN FILE-SIZE                 *
************************************************************
*/
U08 TFS_ExpandFile(TFS_LONG addsize, TFS_HANDLE handl) KC51_REENTRANT
{
    U08 returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > TFS_HANDLES) ||
       (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(TFS_NO_USER);                               // check if file opened
    }
    if(TFS_User[user]->used[handl-1].openEntry->Attr & TFS_ATTR_FILE) {
        returnOk = TFSExpand_File((TFS_DIRFILE OS_HUGE *)(TFS_User[user]->used[handl-1].openEntry), addsize);
    } else {
        returnOk = TFS_NO_FILE;
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*               SET ACTUAL PTR IN OPEN FILE                *
************************************************************
*/
U08 TFS_SeekFile(TFS_LONG offset, TFS_HANDLE handl) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_DATA    OS_HUGE *DataPTR;
    TFS_LONG             DB_offset;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > TFS_HANDLES) ||
       (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
        return(TFS_NO_USER);                               // check if file opened
    }
    if(TFS_User[user]->used[handl-1].openEntry->Attr & TFS_ATTR_FILE) {
        FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->used[handl-1].openEntry);
        if(FilePTR->Sub == NULL) {                         // check if file have a DB-Block
            TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
            return(TFS_NO_DATA);
        }
        DataPTR=(TFS_DATA OS_HUGE *)FilePTR->Sub;
        if(offset != TFS_SEEK_EOF) {                       // calculate offset
            DB_offset = offset;
        } else {
            DB_offset = FilePTR->UsedLength;
        }
        if(DB_offset > FilePTR->UsedLength) {
            TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
            return(TFS_FILE_EOF);                          // offset is larger used-size
        }
        TFS_User[user]->used[handl-1].openOffset = DB_offset;
        while((DB_offset > DataPTR->Length) && (DataPTR->Entry.Next != NULL)) {
            DB_offset -= DataPTR->Length;
            DataPTR    = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);
        }                                                  // calculate DB-Block to access
        if(DB_offset > DataPTR->Length) {
            returnOk = TFS_WRONG_PTR;                      // PTR is larger file-size
        }
    } else {
        returnOk = TFS_NO_FILE;                            // no file opened
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*            SET EOF IN OPEN FILE ON ACTUAL PTR            *
************************************************************
*/
U08 TFS_SetEOF(TFS_HANDLE handl) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *FilePTR = (TFS_DIRFILE *)NULL;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    if((handl == 0) || (handl > TFS_HANDLES) ||
       (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
        returnOk = TFS_NO_USER;
    } else {                                               // check if file opened
        if(TFS_User[user]->used[handl-1].openEntry->Attr & TFS_ATTR_FILE) {
            FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->used[handl-1].openEntry);
            if(FilePTR->Sub == NULL) {                     // check if file have a DB-Block
                returnOk = TFS_NO_DATA;
            }
            if((FilePTR->Entry.Attr & TFS_ATTR_RO) || (TFS_User[user]->used[handl-1].openMode == TFS_ACC_RO)) {
                returnOk = TFS_FILE_RO;                // check if not RO-File or opened in RO-mode
            }
        } else {
            returnOk = TFS_NO_FILE;                    // no file opened
        }
    }
    if(returnOk == TFS_NO_ERR) {
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->UsedLength), (U08 OS_HUGE *)(&TFS_User[user]->used[handl-1].openOffset), sizeof(TFS_LONG));
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex); 
    return(returnOk);
}

/*
************************************************************
*               GET ACTUAL PTR IN OPEN FILE                *
************************************************************
*/
TFS_LONG TFS_TellFile(TFS_HANDLE handl) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_LONG             pos;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk == TFS_NO_ERR) {
        if((handl == 0) || (handl > TFS_HANDLES) ||
           (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
            returnOk = TFS_NO_USER;
        } else {                                           // check if file opened
            if(TFS_User[user]->used[handl-1].openEntry->Attr & TFS_ATTR_FILE) {
                FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->used[handl-1].openEntry);
                if(FilePTR->Sub == NULL) {                 // check if file have a DB-Block
                    returnOk = TFS_NO_DATA;
                }
            } else {
                returnOk = TFS_NO_FILE;                    // no file opened
            }
        }
    }
    TFS_User[user]->ErrNo = returnOk;
    if(returnOk == TFS_NO_ERR) {
        pos = TFS_User[user]->used[handl-1].openOffset;
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(pos);
    }
    TFSMutex_Put(&TFS_mutex);
    return(0);
}

/*
************************************************************
*                         READ FILE                        *
************************************************************
*/
TFS_LONG TFS_ReadFile(U08 OS_HUGE *dest, TFS_LONG size, TFS_HANDLE handl) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_DATA    OS_HUGE *DataPTR;
    U08         OS_HUGE *src_PTR;
    TFS_LONG             i, j, DB_offset, offset;
    U08                  user;

    if(TFS_NO_ERR != TFSCheckUser(&user)) {
        return(0);                                         // no correct user
    }
    if((dest >= (U08 OS_HUGE *)(TFS_START)) && (dest < (U08 OS_HUGE *)(TFS_START) + TFS_SIZE)) {
        TFS_User[user]->ErrNo = TFS_WR_PTR;
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // "dest" can't be in TFS-Area
    }
    if((handl == 0) || (handl > TFS_HANDLES) ||
       (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
        TFS_User[user]->ErrNo = TFS_NO_USER;
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // check if file opened
    }
    FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->used[handl-1].openEntry);
    if(FilePTR->Entry.Attr & TFS_ATTR_FILE) {
        if((FilePTR->Entry.Attr & TFS_ATTR_WO) || (TFS_User[user]->used[handl-1].openMode == TFS_ACC_WO)) {
            TFS_User[user]->ErrNo = TFS_FILE_WO;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // check file to write-access
        }
        if(FilePTR->Sub == NULL) {                         // check if file have a DB-Block
            TFS_User[user]->ErrNo = TFS_NO_DATA;
            TFSMutex_Put(&TFS_mutex);
            return(0);
        }
        DataPTR = (TFS_DATA OS_HUGE *)FilePTR->Sub;
        offset = DB_offset = TFS_User[user]->used[handl-1].openOffset;
        if(DB_offset > FilePTR->UsedLength) {
            TFS_User[user]->ErrNo = TFS_FILE_EOF;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // PTR is larger used-size
        }
        while(DB_offset > DataPTR->Length && DataPTR->Entry.Next != NULL) {
            DB_offset -= DataPTR->Length;
            DataPTR    = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);
        }                                                  // calculate DB-Block to access
        if(DB_offset > DataPTR->Length) {
            TFS_User[user]->ErrNo = TFS_WRONG_PTR;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // PTR is larger file-size
        }
        j = 0;
        while(size > 0 ) {
            i = 0;                                         // read data-bytes
            src_PTR = (U08 OS_HUGE *)(DataPTR) + sizeof(TFS_DATA) + DB_offset;
            while((i < size) && (DB_offset+i < DataPTR->Length) && ((offset+j+i) < FilePTR->UsedLength)) {
                dest[j+i] = src_PTR[i]; i++;
            }
            size -= i; j += i;                             // set actual offset
            if(size > 0) {                                 // if next DB-Block needed
                if(((offset+j) >= FilePTR->UsedLength) || (DataPTR->Entry.Next == NULL)) {
                    offset += j;
                    TFS_User[user]->ErrNo = TFS_FILE_EOF;  // EndOfFile is found
                    TFS_User[user]->used[handl-1].openOffset = offset;
                    TFSMutex_Put(&TFS_mutex);
                    return(j);
                }
                DataPTR = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);
                DB_offset = 0;
            }
        }
        offset += j;
        TFS_User[user]->used[handl-1].openOffset = offset;
    } else {
        TFS_User[user]->ErrNo = TFS_NO_FILE;
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // no file opened
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(j);
}

/*
************************************************************
*                         WRITE FILE                       *
************************************************************
*/
TFS_LONG TFS_WriteFile(U08 OS_HUGE *src, TFS_LONG size, TFS_HANDLE handl) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_DATA    OS_HUGE *DataPTR;
    TFS_LONG             i, j, DB_offset, offset;
    U08                  user;

    if(TFS_NO_ERR != TFSCheckUser(&user)) {
        return(0);
    }
    if((src >= (U08 OS_HUGE *)(TFS_START)) && (src < (U08 OS_HUGE *)(TFS_START)+TFS_SIZE)) {
        TFS_User[user]->ErrNo = TFS_WR_PTR;
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // "src" can't be in TFS-Area
    }
    if((handl == 0) || (handl > TFS_HANDLES) ||
       (TFS_User[user]->used[handl-1].openEntry == (TFS_ENTRY OS_HUGE *)NULL)) {
        TFS_User[user]->ErrNo = TFS_NO_USER;
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // check if file opened
    }
    FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->used[handl-1].openEntry);
    if(FilePTR->Entry.Attr & TFS_ATTR_FILE) {
        if((FilePTR->Entry.Attr & TFS_ATTR_RO) || (TFS_User[user]->used[handl-1].openMode == TFS_ACC_RO)) {
            TFS_User[user]->ErrNo = TFS_FILE_RO;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // check file to write-access
        }
        if(FilePTR->Sub == NULL) {
            TFS_User[user]->ErrNo = TFS_NO_DATA;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // check if file have a DB-Block
        }
        DataPTR = (TFS_DATA OS_HUGE *)(FilePTR->Sub);
        offset = DB_offset = TFS_User[user]->used[handl-1].openOffset;
        if(DB_offset > FilePTR->UsedLength) {
            TFS_User[user]->ErrNo = TFS_FILE_EOF;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // PTR is larger used-size
        }
        while((DB_offset > DataPTR->Length) && (DataPTR->Entry.Next != NULL)) {
            DB_offset -= DataPTR->Length;
            DataPTR    = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);
        }                                                  // calculate DB-Block to access
        if(DB_offset > DataPTR->Length) {
            TFS_User[user]->ErrNo = TFS_WRONG_PTR;
            TFSMutex_Put(&TFS_mutex);
            return(0);                                     // PTR is larger file-size
        }
        j = 0;
        while(size > 0 ) {                                 // write data-bytes
            if(DB_offset + size > DataPTR->Length) {
                i = DataPTR->Length - DB_offset;
            } else {                                       // for every needed DB-Block
                i = size;
            }
            TFS_User[user]->ErrNo = TFSPort_Write((U08 OS_HUGE *)(DataPTR)+sizeof(TFS_DATA)+DB_offset, &src[j], i);
            if(TFS_User[user]->ErrNo != TFS_NO_ERR) {
                TFSMutex_Put(&TFS_mutex);
                return(j);
            }
            size -= i; j += i;
            if(size > 0) {                                 // if next DB-Block needed
                if(DataPTR->Entry.Next == NULL) {
                    offset += j;
                    TFS_User[user]->ErrNo = TFS_FILE_EOF;
                    TFS_User[user]->used[handl-1].openOffset=offset;
                    if(FilePTR->UsedLength < offset) {
                        TFSPort_Write((U08 OS_HUGE *)(&FilePTR->UsedLength), (U08 OS_HUGE *)(&offset), sizeof(TFS_LONG));
                    }
                    TFSMutex_Put(&TFS_mutex);
                    return(j);
                }
                DataPTR = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);
                DB_offset = 0;
            }
        }
        offset += j;
        TFS_User[user]->used[handl-1].openOffset = offset; // set actual offset
        if(FilePTR->UsedLength < offset) {
            if((TFS_User[user]->ErrNo = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->UsedLength), (U08 OS_HUGE *)(&offset), sizeof(TFS_LONG))) != TFS_NO_ERR) {
                TFSMutex_Put(&TFS_mutex);
                return(j);
            }
        }
    } else {
        TFS_User[user]->ErrNo = TFS_NO_FILE;
        TFSMutex_Put(&TFS_mutex);
        return(0);                                         // no file opened
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(j);
}

/*
************************************************************
*                         WRITE FILE PLUS                  *
************************************************************
*/
TFS_LONG TFS_WriteFileE(U08 OS_HUGE *src, TFS_LONG size, TFS_HANDLE handl) KC51_REENTRANT
{
    TFS_LONG written;
    U08      user;

    user = TFSGetUserByID();
    if(user == 0xFF) {
        written = 0;
    } else {
        written = TFS_WriteFile(src, size, handl);
        if((written != size) && (TFS_User[user]->ErrNo == TFS_FILE_EOF)) {
            TFS_User[user]->ErrNo = TFS_ExpandFile(size - written, handl);
            if(TFS_User[user]->ErrNo == TFS_NO_ERR) {
                written += TFS_WriteFile(&src[written], size - written, handl);
            }
        }
    }
    return(written);
}

/*
************************************************************
*                         CREATE LINK                      *
************************************************************
*/
U08 TFS_CreateLink(TFS_PATHNAME OS_HUGE *OrgName, TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_ENTRY   OS_HUGE *currPTR;
    TFS_DIRFILE OS_HUGE *orgPTR;
    TFS_DIRFILE OS_HUGE *linkPTR;
    U08                  returnOk, i, user;
    TFS_LONG             Pos;
    TFS_ATTR             a;

    Pos = 0;
    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR){
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go into org-PATH
    returnOk = TFSChange_Entry(OrgName, 1);                // go into * / stay on link-*
    if(returnOk == TFS_NO_ERR) {
        orgPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
        TFS_User[user]->currentEntry = currPTR;            // restore saved position
        returnOk = TFSChange_Entry(Name, 0);               // go up to last part of path
        if(returnOk == TFS_NO_ERR) {
            Pos = TFSstrrpos(Name, '/');
            if(Pos == (TFS_LONG)(0-1)) {
                Pos = 0;
            } else {
                Pos++;
            }
            if((Pos < TFSstrlen(Name)) && (TFSstrlen(Name)-Pos < TFS_Name_SIZE)) {
                i = orgPTR->LinkCnt;
                if(i < 255) {
                    i++;
                    returnOk = TFSPort_Write((U08 OS_HUGE *)(&orgPTR->LinkCnt), (U08 OS_HUGE *)(&i), sizeof(U08));
                    if(returnOk == TFS_NO_ERR) {
                        a = TFS_ATTR_LINK | (orgPTR->Entry.Attr & (TFS_ATTR_DIR | TFS_ATTR_FILE));
                        returnOk = TFSCreate_Entry(&Name[Pos], a, 0);
                        if(returnOk == TFS_NO_ERR) {
                            returnOk = TFSChange_Entry(&Name[Pos], 1); // go into * / stay on link-*
                            if(returnOk == TFS_NO_ERR) {
                                linkPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
                                returnOk = TFSPort_Write((U08 OS_HUGE *)(&linkPTR->Sub), (U08 OS_HUGE *)(&orgPTR), sizeof(U08 OS_HUGE *));
                                if(returnOk != TFS_NO_ERR) {
#if (TFSDBLLINKED==1)
                                    TFSDelete_Entry(TFS_User[user]->currentEntry);
#else
                                    TFSDelete_Entry(TFS_User[user]->currentEntry, TFS_User[user]->prevEntry);
#endif
                                }
                            }
                        }
                    }
                } else {
                    returnOk = TFS_MAX_LINK;
                }
            } else {
                returnOk = TFS_PATH_ERR;                   // no valid link-name given
            }
        }
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         REMOVE LINK                      *
************************************************************
*/
U08 TFS_RemoveLink(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *linkPTR;
    TFS_ENTRY   OS_HUGE *currPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go to link (not into link)
    returnOk = TFSChange_Entry(Name, 1);                   // go into * / stay on link-*
    if(returnOk == TFS_NO_ERR) {
        linkPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
        if(linkPTR->LinkCnt == 0) {
            if(linkPTR->Sub != NULL) {
                linkPTR = (TFS_DIRFILE OS_HUGE *)(linkPTR->Sub);
                i = linkPTR->LinkCnt;
                if(i > 0) {
                    i--;
                    returnOk = TFSPort_Write((U08 OS_HUGE *)(&linkPTR->LinkCnt), (U08 OS_HUGE *)(&i), sizeof(U08));
                    if(returnOk == TFS_NO_ERR) {
#if (TFSDBLLINKED==1)
                        returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry);
#else
                        returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry, TFS_User[user]->prevEntry);
#endif
                    }
                } else {
                    returnOk = TFS_LINK_ERR;
                }
            } else {
                returnOk = TFS_NO_LINK;                    // no link-PTR in Link-Entry
            }
        } else {
            returnOk = TFS_LINKED;                         // link is linked
        }
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                    READ LINKED ENTRY-NAME                *
************************************************************
*/
U08 TFS_ReadLink(TFS_PATHNAME OS_HUGE *Name, U08 OS_HUGE *LinkName) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *linkPTR;
    TFS_ENTRY   OS_HUGE *currPTR;
    U08                  returnOk ,i , user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go into link
    returnOk = TFSChange_Entry(Name, 4);                   // go 1 level into link-*
    if(returnOk == TFS_NO_ERR) {
        linkPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
        i = 0;
        do {
            LinkName[i] = linkPTR->Name[i]; i++;
        } while(linkPTR->Name[i] != 0);
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         CREATE DIR                       *
************************************************************
*/
U08 TFS_CreateDir(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE   *currPTR;
    U08                  returnOk, user;
    TFS_LONG             Pos;

    Pos = 0;                                               // correct attributes
    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 0);                   // go up to last part of path
    if(returnOk == TFS_NO_ERR) {                           // create Entry
        Pos = TFSstrrpos(Name, '/');
        if(Pos == (TFS_LONG)(0-1)) {
            Pos = 0;
        } else {
            Pos++;
        }
        if((Pos < TFSstrlen(Name)) && (TFSstrlen(Name)-Pos < TFS_Name_SIZE)) { // create Entry
            returnOk = TFSCreate_Entry(&Name[Pos], TFS_ATTR_DIR, 0);
        } else {
            returnOk = TFS_PATH_ERR;
        }
    }
    TFS_User[user]->currentEntry = currPTR;
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         REMOVE DIR                       *
************************************************************
*/
U08 TFS_RemoveDir(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *dirPTR;
    TFS_ENTRY   OS_HUGE *currPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 2);                   // go into dir / stay on link-dir
    if(returnOk == TFS_NO_ERR) {                           // Delete Entry if no sub-elements and not the ROOT
        dirPTR = (TFS_DIRFILE OS_HUGE *)TFS_User[user]->currentEntry;
        if(dirPTR->LinkCnt == 0) {
            if((((dirPTR->Sub != NULL) && (dirPTR->Entry.Attr & TFS_ATTR_LINK)) ||
                (dirPTR->Sub == NULL)) && (dirPTR->Entry.PrevLevel != NULL)) {
#if (TFS_SubROOT==1)
                for(i = 0; i < TFS_MAX_USER; i++) {
                    if(TFS_User[i] != (TFS_USER OS_HUGE *)NULL) {
                        if(TFS_User[i]->SubROOT == (TFS_ENTRY OS_HUGE *)dirPTR) {
                            TFS_User[user]->currentEntry = currPTR;     // restore saved position
                            TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
                            return(TFS_SUB_USER);
                        }
                    }
                }
#endif
                if((dirPTR->Entry.Attr & TFS_ATTR_LINK)) {
                    dirPTR = (TFS_DIRFILE OS_HUGE *)(dirPTR->Sub);
                    i = dirPTR->LinkCnt;
                    if(i > 0) {
                        i--;
                        returnOk = TFSPort_Write((U08 OS_HUGE *)(&dirPTR->LinkCnt), (U08 OS_HUGE *)(&i), sizeof(U08));
                        if(returnOk != TFS_NO_ERR) {
                            TFS_User[user]->currentEntry = currPTR;  // restore saved position
                            TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
                            return(returnOk);
                        }
                    } else {
                        TFS_User[user]->currentEntry=currPTR; // restore saved position
                        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
                        return(TFS_LINK_ERR);
                    }
                }
                for(i = 0; i < TFS_MAX_USER; i++) {
                    if(TFS_User[i] != (TFS_USER OS_HUGE *)NULL) {
                        if(i != user) {
                            if(TFS_User[i]->currentEntry == TFS_User[user]->currentEntry) {
                                TFS_User[i]->currentEntry = TFS_User[user]->currentEntry->PrevLevel;
                            }
#if (TFS_TempDIR==1)
                            if(TFS_User[i]->TempDir == TFS_User[user]->currentEntry) {
                                TFS_User[i]->TempDir = TFS_User[user]->currentEntry->PrevLevel;
                            }
#endif
                        } else {
                            if(currPTR == TFS_User[user]->currentEntry) {
                                currPTR = TFS_User[user]->currentEntry->PrevLevel;
                            }
#if (TFS_TempDIR==1)
                            if(currPTR == TFS_User[user]->TempDir) {
                                currPTR = TFS_User[user]->TempDir->PrevLevel;
                            }
#endif
                        }
                    }
                }
#if (TFSDBLLINKED==1)
                returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry);
#else
                returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry, TFS_User[user]->prevEntry);
#endif
            } else {
                returnOk = TFS_NOT_EMPTY;                  // DIR not empty
            }
        } else {
            returnOk = TFS_LINKED;                         // entry is linked
        }
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         RENAME DIR                       *
************************************************************
*/
U08 TFS_RenameDir(TFS_PATHNAME OS_HUGE *OldName, TFS_PATHNAME OS_HUGE *NewName) KC51_REENTRANT
{
    TFS_ENTRY   OS_HUGE *orgPTR;
    TFS_DIRFILE OS_HUGE *DirPTR;
    TFS_DIRFILE OS_HUGE *SDirPTR;
    TFS_DIRFILE OS_HUGE *EntryPTR;
    TFS_LONG             Pos, i;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    orgPTR = TFS_User[user]->currentEntry;                 // save actual position and go into PATH
    returnOk = TFSChange_Entry(OldName, 2);                // go into dir / stay on link-dir
    if(returnOk == TFS_NO_ERR) {                           // if valid old name
        DirPTR  = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
        SDirPTR = (TFS_DIRFILE OS_HUGE *)(DirPTR->Entry.PrevLevel);
        Pos = TFSstrrpos(NewName,'/');                     // search last '/' in PATH or full name and check
        if(Pos == (TFS_LONG)(0-1)) {
            Pos = 0;
        } else {
            Pos++;                                         // update name if NewName not exist
        }
        if((Pos < TFSstrlen(NewName)) && (TFSstrlen(NewName)-Pos < TFS_Name_SIZE)) {
            EntryPTR = (TFS_DIRFILE OS_HUGE *)(SDirPTR->Sub); // get first entry and check if name exist
            while((0 != (i = TFSstrcmp(&NewName[Pos], (TFS_PATHNAME OS_HUGE *)EntryPTR->Name))) && (NULL != EntryPTR->Entry.Next)) {
                EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
            }
            if(i == 0) {                                   // up to end of level
                returnOk = TFS_NAME_EXIST;                 // NewName exist
            } else {
                returnOk = TFSPort_Write((U08 OS_HUGE *)DirPTR->Name, (U08 OS_HUGE *)&NewName[Pos], TFSstrlen(&NewName[Pos])+1);
            }
        } else {
            returnOk = TFS_PATH_ERR;                       // error in new-name
        }
    } else {
        returnOk = TFS_PATH_ERR;                           // no existing PATH given
    }
    TFS_User[user]->currentEntry = orgPTR;                 // restore actual position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         CREATE FILE                      *
************************************************************
*/
U08 TFS_CreateFile(TFS_PATHNAME OS_HUGE *Name, TFS_ATTR Attribs, TFS_LONG size) KC51_REENTRANT
{
    TFS_ENTRY OS_HUGE   *currPTR;
    TFS_LONG             Pos;
    TFS_ATTR             a;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    a  = Attribs & (TFS_ATTR_RO | TFS_ATTR_WO |TFS_ATTR_SYS); 
    a |= TFS_ATTR_FILE;
    Pos = 0;
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go into PATH
    returnOk = TFSChange_Entry(Name, 0);                   // go up to last part of path
    if(returnOk == TFS_NO_ERR) {                           // if valid file-name - create entry, go into and create DB-Block
        Pos = TFSstrrpos(Name, '/');
        if(Pos == (U16)(0-1)) {
            Pos = 0;
        } else {
            Pos++;                                         // create Entry
        }
        if((Pos < TFSstrlen(Name)) && (TFSstrlen(Name)-Pos < TFS_Name_SIZE)) {
            returnOk = TFSCreate_Entry(&Name[Pos], a, 0);
            if(returnOk == TFS_NO_ERR) {
                returnOk = TFSChange_Entry(&Name[Pos], 3); // go into file / stay on link-file
                if(returnOk == TFS_NO_ERR) {
                    a = TFS_ATTR_DB;                       // attribs only in header
                    returnOk = TFSCreate_Entry((TFS_PATHNAME OS_HUGE *)NULL, a, size);
                    if(returnOk != TFS_NO_ERR) {           // delete entry if was error on create DB-Block
#if (TFSDBLLINKED==1)
                        returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry);
#else
                        returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry, TFS_User[user]->prevEntry);
#endif
                    }
                }
            }
        } else {
            returnOk = TFS_NO_FILE;                        // no valid file-name given
        }
    }                                                      // restore saved position
    TFS_User[user]->currentEntry = currPTR;
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         REMOVE FILE                      *
************************************************************
*/
U08 TFS_RemoveFile(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *FilePTR;
    TFS_ENTRY   OS_HUGE *currPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position and go into file
    returnOk = TFSChange_Entry(Name, 3);                   // go into file / stay on link-file
    if(returnOk == TFS_NO_ERR) {
        returnOk = TFSFileOpend(TFS_User[user]->currentEntry);
        if(returnOk == TFS_NO_ERR) {
            FilePTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
            if(FilePTR->LinkCnt == 0) {
                if((FilePTR->Entry.Attr & TFS_ATTR_LINK)) {
                    FilePTR = (TFS_DIRFILE OS_HUGE *)(FilePTR->Sub);
                    i = FilePTR->LinkCnt;
                    if(i > 0) {
                        i--;
                        returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->LinkCnt), (U08 OS_HUGE *)(&i), sizeof(U08));
                        if(returnOk == TFS_NO_ERR) {
#if (TFSDBLLINKED==1)
                            returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry);
#else
                            returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry, TFS_User[user]->prevEntry);
#endif
                        }
                    } else {
                        returnOk = TFS_LINK_ERR;
                    }
                } else {
                    while((FilePTR->Sub != NULL) && (returnOk == TFS_NO_ERR)) {
#if (TFSDBLLINKED==1)
                        returnOk = TFSDelete_Entry(FilePTR->Sub);
#else
                        returnOk = TFSDelete_Entry(FilePTR->Sub, NULL);
#endif
                    }
                    if(FilePTR->Sub == NULL) {             // delete all DB-Blocks
#if (TFSDBLLINKED==1)
                        returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry);
#else
                        returnOk = TFSDelete_Entry(TFS_User[user]->currentEntry, TFS_User[user]->prevEntry);
#endif
                    }
                }
            } else {
                returnOk = TFS_LINKED;                     // entry is linked
            }
        }
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                         CHANGE DIR                       *
************************************************************
*/
U08 TFS_ChangeDir(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    U08 returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSChange_Entry(Name, 5);                   // go into dir / into linked dir
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

#if (TFS_TempDIR==1)
/*
************************************************************
*                    CHANGE DIR TEMPORARY                  *
************************************************************
*/
U08 TFS_ChangeDirTemp(TFS_PATHNAME OS_HUGE *Name) KC51_REENTRANT
{
    U08   returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    if(TFS_User[user]->TempDir != NULL) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(TFS_TMP_DIR);
    }
    TFS_User[user]->TempDir = TFS_User[user]->currentEntry;  // save current dir
    if(TFS_User[user]->TempDir->Attr & TFS_ATTR_FILE) {
        TFS_User[user]->TempDir = TFS_User[user]->TempDir->PrevLevel;
    }
    returnOk = TFSChange_Entry(Name, 5);                   // go into dir / into linked dir
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*                 RESTORE DIR FROM TEMPORARY               *
************************************************************
*/
U08 TFS_BackDirTemp(void) KC51_REENTRANT
{
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    if(TFS_User[user]->TempDir != NULL) {                  // restore saved current dir
        TFS_User[user]->currentEntry = TFS_User[user]->TempDir;
        TFS_User[user]->TempDir      = NULL;
        return(TFS_NO_ERR);
    }
    return(TFS_TMP_DIR);
}
#endif

/*
************************************************************
*                      GET CURRENT DIR                     *
************************************************************
*/
U08 TFS_GetCurrentDir(U08 OS_HUGE *Name) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *dirPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    dirPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
    if(dirPTR->Entry.Attr & TFS_ATTR_FILE) {
        dirPTR = (TFS_DIRFILE OS_HUGE *)(dirPTR->Entry.PrevLevel);
    }
    i=0;                                                   // if actual in file - get name of DIR one back
    do {
        Name[i] = dirPTR->Name[i]; i++;
    } while((i < TFS_Name_SIZE) && (Name[i-1] != 0));
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*         GET FULL PATH (internal - called recursiv)       *
************************************************************
*/
static TFS_LONG TFSGetFullPath(TFS_DIRFILE OS_HUGE *actPTR, U08 OS_HUGE *Path, TFS_LONG maxlen, U08 user) KC51_REENTRANT
{
    TFS_LONG     i, offs;

    offs = 0;
#if (TFS_SubROOT==1)
    if((actPTR != (TFS_DIRFILE OS_HUGE *)TFS_User[user]->SubROOT) && (actPTR != (TFS_DIRFILE OS_HUGE *)(ROOT_PTR))) {
#else
    if(actPTR != (TFS_DIRFILE OS_HUGE *)(ROOT_PTR)) {
#endif
        offs = TFSGetFullPath((TFS_DIRFILE OS_HUGE *)actPTR->Entry.PrevLevel, Path, maxlen, user);    // if not ROOT / SubROOT - go one dir back
    }
    i = 0;
    for(i = 0; (i < TFS_Name_SIZE) && ((offs + i) < maxlen) && (actPTR->Name[i] != 0); i++) {
        Path[offs + i] = actPTR->Name[i];
    }
    if(((offs + i) < maxlen) && (offs > 0)) {
        Path[offs + i++] = '/';
    }
    return(offs + i);
}

/*
************************************************************
*                     GET CURRENT PATH                     *
************************************************************
*/
U08 TFS_GetCurrentPath(U08 OS_HUGE *Path, TFS_LONG maxlen) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *dirPTR;
    TFS_LONG             len;
    U08                  returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    dirPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
    if(dirPTR->Entry.Attr & TFS_ATTR_FILE) {               // if actual in file - go one back to DIR
        dirPTR = (TFS_DIRFILE OS_HUGE *)(dirPTR->Entry.PrevLevel);
    }
    len = TFSGetFullPath(dirPTR, Path, maxlen, user);      // start recursive Path-notice
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    if(len >= maxlen) {
        Path[maxlen -1] = 0;
        return(TFS_MEM_OVF);
    }
    Path[len] = 0;                                         // mark end-of-string
    return(TFS_NO_ERR);
}

/*
************************************************************
*                    GET FIRST ENTRY (internal)            *
************************************************************
*/
static U08 TFSGetFirstEntry(TFS_DIRFILE OS_HUGE **Entry) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  user;

    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);
    }
    EntryPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
    if(EntryPTR->Entry.Attr & TFS_ATTR_FILE) {
        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.PrevLevel);
    }
    if(EntryPTR->Sub == NULL) {                            // if no Entry exist
        return(TFS_NOT_EXIST);                             // no entries in actual DIR
    }
    EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Sub);     // dont show SYS- Dir/Files
    while((EntryPTR != NULL) && (EntryPTR->Entry.Attr & TFS_ATTR_SYS)) {
        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
    }
    TFS_User[user]->getEntry = (TFS_ENTRY OS_HUGE *)(EntryPTR);
    if(EntryPTR == NULL) {
        return(TFS_NOT_EXIST);                             // no additional entries in actual DIR
    }
    *Entry = EntryPTR;
    return(TFS_NO_ERR);
}

/*
************************************************************
*                    GET NEXT ENTRY (internal)             *
************************************************************
*/
static U08 TFSGetNextEntry(TFS_DIRFILE OS_HUGE **Entry) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  user;

    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);
    }
    if((TFS_User[user]->getEntry == NULL) || (TFS_User[user]->getEntry->PrevLevel != TFS_User[user]->currentEntry)) {
        return(TFS_PATH_ERR);
    }
    EntryPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->getEntry->Next);
    while((EntryPTR != NULL) && (EntryPTR->Entry.Attr & TFS_ATTR_SYS)) {  // dont show SYS- Dir/Files
        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
    }
    if(EntryPTR == NULL) {
        return(TFS_NOT_EXIST);                             // no additional entries in actual DIR
    }
    TFS_User[user]->getEntry = (TFS_ENTRY OS_HUGE *)(EntryPTR);
    *Entry = EntryPTR;
    return(TFS_NO_ERR);
}

/*
************************************************************
*                    GET FIRST ENTRY-NAME                  *
************************************************************
*/
U08 TFS_GetFirstName(U08 OS_HUGE *Name) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSGetFirstEntry(&EntryPTR);
    if(returnOk != TFS_NO_ERR) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(returnOk);
    }
    i=0;
    do {
        Name[i] = EntryPTR->Name[i]; i++;
    } while((i < TFS_Name_SIZE) && (Name[i-1] != 0));
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(TFS_NO_ERR);
}

/*
************************************************************
*                    GET NEXT ENTRY-NAME                   *
************************************************************
*/
U08 TFS_GetNextName(U08 OS_HUGE *Name) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSGetNextEntry(&EntryPTR);
    if(returnOk != TFS_NO_ERR) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(returnOk);
    }
    i=0;
    do {
        Name[i] = EntryPTR->Name[i]; i++;
    } while((i < TFS_Name_SIZE) && (Name[i-1] != 0));
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(TFS_NO_ERR);
}

/*
************************************************************
*                    GET FIRST ENTRY-INFOS                 *
************************************************************
*/
U08 TFS_GetFirst(TFS_GET OS_HUGE *get) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSGetFirstEntry(&EntryPTR);
    if(returnOk != TFS_NO_ERR) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(returnOk);
    }
    i=0;
    do {
        get->Name[i] = EntryPTR->Name[i]; i++;
    } while((i < TFS_Name_SIZE) && (get->Name[i-1] != 0));
    get->Attr   = EntryPTR->Entry.Attr & (TFS_ATTR_DIR | TFS_ATTR_FILE | TFS_ATTR_LINK | TFS_ATTR_WO | TFS_ATTR_RO);
    get->Length = EntryPTR->UsedLength;
    if(EntryPTR->LinkCnt != 0) {
        get->Linked = TRUE;
    } else {
        get->Linked = FALSE;
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(TFS_NO_ERR);
}

/*
************************************************************
*                    GET NEXT ENTRY-INFOS                  *
************************************************************
*/
U08 TFS_GetNext(TFS_GET OS_HUGE *get) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    returnOk = TFSGetNextEntry(&EntryPTR);
    if(returnOk != TFS_NO_ERR) {
        TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
        return(returnOk);
    }
    i=0;
    do {
        get->Name[i] = EntryPTR->Name[i]; i++;
    } while((i < TFS_Name_SIZE) && (get->Name[i-1] != 0));
    get->Attr   = EntryPTR->Entry.Attr & (TFS_ATTR_DIR | TFS_ATTR_FILE | TFS_ATTR_LINK | TFS_ATTR_WO | TFS_ATTR_RO);
    get->Length = EntryPTR->UsedLength;
    if(EntryPTR->LinkCnt != 0) {
        get->Linked = TRUE;
    } else {
        get->Linked = FALSE;
    }
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(TFS_NO_ERR);
}

/*
************************************************************
*      set the current getEntry  (for recurive reasons)    *
************************************************************
*/
static U08 _TFSgetEntry_set(TFS_ENTRY OS_HUGE *EntryPTR)
{
    U08                returnOk, user;

    if(EntryPTR == NULL) {
        return(TFS_NOT_EXIST);
    }
    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    TFS_User[user]->getEntry = EntryPTR;
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(TFS_NO_ERR);
}

/*
************************************************************
*      get the current getEntry  (for recurive reasons)    *
************************************************************
*/
static TFS_ENTRY OS_HUGE *_TFSgetEntry_give(void)
{
    TFS_ENTRY OS_HUGE *EntryPTR;
    U08                returnOk, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        TFS_User[user]->ErrNo = returnOk;
        return(NULL);
    }
    EntryPTR = TFS_User[user]->getEntry;
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(EntryPTR);
}

/*
************************************************************
*                    GET NAMED ENTRY-INFOS                 *
************************************************************
*/
U08 TFS_GetEntry(TFS_PATHNAME OS_HUGE *Name, TFS_GET OS_HUGE *get) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    TFS_ENTRY   OS_HUGE *currPTR;
    U08                  returnOk, i, user;

    returnOk = TFSCheckUser(&user);
    if(returnOk != TFS_NO_ERR) {
        return(returnOk);
    }
    currPTR = TFS_User[user]->currentEntry;                // save actual position
    returnOk = TFSChange_Entry(Name, 1);                   // go into * / stay on link-*
    if(returnOk == TFS_NO_ERR) {
        EntryPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry);
        i=0;
        do {
            get->Name[i] = EntryPTR->Name[i]; i++;
        } while((i < TFS_Name_SIZE) && (get->Name[i-1] != 0));
        get->Attr   = EntryPTR->Entry.Attr & (TFS_ATTR_DIR | TFS_ATTR_FILE | TFS_ATTR_LINK | TFS_ATTR_WO | TFS_ATTR_RO);
        get->Length = EntryPTR->UsedLength;
        if(EntryPTR->LinkCnt != 0) {
            get->Linked = TRUE;
        } else {
            get->Linked = FALSE;
        }
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved position
    TFS_User[user]->ErrNo = TFSMutex_Put(&TFS_mutex);
    return(returnOk);
}

/*
************************************************************
*              CHECK IF FILE IS OPEND (internal)           *
************************************************************
*/
static U08 TFSFileOpend(TFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT
{
    U08 i, j;

    for(i = 0; i < TFS_MAX_USER; i++) {
        if(TFS_User[i] != (TFS_USER OS_HUGE *)NULL) {
            for(j = 0; j < TFS_HANDLES; j++) {
                if(TFS_User[i]->used[j].openEntry == EntryPTR) {
                    return(TFS_FILE_OPEN);
                }
            }
        }
    }
    return(TFS_NO_ERR);
}

/*
************************************************************
*                   CHANGE ENTRY (internal)                *
*                                                          *
*   mode:   0 - go up to last part of path                 *
*           1 - go into *    / stay on link-*              *
*           2 - go into dir  / stay on link-dir            *
*           3 - go into file / stay on link-file           *
*           4 - go             1 level into link-*         *
*           5 - go into dir  / into linked dir             *
*           6 - go into file / into linked file            *
*                                                          *
************************************************************
*/
static U08 TFSChange_Entry(TFS_PATHNAME OS_HUGE *Name, U08 mode) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    TFS_ENTRY   OS_HUGE *currPTR;
    TFS_LONG             Pos, Pos1;
    U08                  returnOk, i, user;

    Pos = 0;
    returnOk = TFS_NO_ERR;
    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);
    }
    currPTR = TFS_User[user]->currentEntry;
    while(Name[Pos] == '/') {                              // is upto ROOT (for all root-symbols at front)
#if (TFS_SubROOT==1)
      if(TFS_User[user]->SubROOT != NULL) {
            currPTR = TFS_User[user]->SubROOT;
        } else {
            currPTR = (TFS_ENTRY OS_HUGE *)(ROOT_PTR);
        }
#else
        currPTR = (TFS_ENTRY OS_HUGE *)(ROOT_PTR);
#endif
        Pos++;
    }
    while(Pos < TFSstrlen(Name)) {                         // up to end of PATH
        if((Name[Pos] == '.') && (Name[Pos+1] == '.') &&
           ((Name[Pos+2] == '/') || (Pos+2 == TFSstrlen(Name)))) {
#if (TFS_SubROOT==1)
            if((currPTR != TFS_User[user]->SubROOT) && (currPTR != (TFS_ENTRY OS_HUGE *)(ROOT_PTR))) {
#else
            if(currPTR != (TFS_ENTRY OS_HUGE *)(ROOT_PTR)) {
#endif
                currPTR = currPTR->PrevLevel;              // go one dir back
            } else {
                return(TFS_PATH_ERR);                      // still in ROOT / SubROOT
            }
            Pos += 3;                                      // if one back
        } else {                                           // search end of next part of PATH
            do {
                Pos1 = TFSstrpos(&Name[Pos],'/');
                if(Pos1 == 0) {
                    Pos++;
                }
            } while(Pos1 == 0);
            if((Pos1 != (TFS_LONG)(0-1)) || (mode > 0)) {
                if(Pos1 == (TFS_LONG)(0-1)) {
                    Pos1 = TFSstrlen(Name)-Pos;
                }
                if(Pos1 < TFS_Name_SIZE) {                 // if name correct - search entry and go into
                    EntryPTR = (TFS_DIRFILE OS_HUGE *)(currPTR);
                    if(EntryPTR->Sub != NULL) {
                        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Sub);
#if (!(TFSDBLLINKED==1))
                        TFS_User[user]->prevEntry = NULL;
#endif
                        i = 1;
                        while(((TFSstrlen((TFS_PATHNAME OS_HUGE *)EntryPTR->Name) != Pos1) ||
                              ((i = TFSstrncmp(&Name[Pos], (TFS_PATHNAME OS_HUGE *)EntryPTR->Name, Pos1)) != 0)) &&
                              (EntryPTR->Entry.Next != NULL)) {
#if (!(TFSDBLLINKED==1))
                            TFS_User[user]->prevEntry = (TFS_ENTRY OS_HUGE *)(EntryPTR);
#endif
                            EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
                        }
                        if((i != 0) && (EntryPTR->Entry.Next == NULL)) {
                            return(TFS_NOT_EXIST);         // entry not exist
                        }
                        Pos += Pos1;
                        if(Pos == TFSstrlen(Name)) {       // if last part of path
                            if((mode == 1) ||
                            ((mode == 2) && (EntryPTR->Entry.Attr & TFS_ATTR_DIR)) ||
                            ((mode == 3) && (EntryPTR->Entry.Attr & TFS_ATTR_FILE)))
                                currPTR = (TFS_ENTRY OS_HUGE *)(EntryPTR);
                            else {
                                if(((mode == 5) && (EntryPTR->Entry.Attr & TFS_ATTR_DIR)) ||
                                ((mode == 6) && (EntryPTR->Entry.Attr & TFS_ATTR_FILE))) {
                                    while(EntryPTR->Entry.Attr & TFS_ATTR_LINK) {
                                        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Sub);
                                        if(!(EntryPTR) || (EntryPTR->LinkCnt == 0)) {
                                            return(TFS_NO_LINK);
                                        }
                                    }
                                    currPTR = (TFS_ENTRY OS_HUGE *)(EntryPTR);
                                } else {
                                    if((mode == 4) && (EntryPTR->Entry.Attr & TFS_ATTR_LINK)) {
                                        EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Sub);
                                        if(!(EntryPTR) || (EntryPTR->LinkCnt == 0)) {
                                            return(TFS_NO_LINK);
                                        }
                                        currPTR = (TFS_ENTRY OS_HUGE *)(EntryPTR);
                                    } else {
                                        Pos = TFSstrlen(Name);
                                        returnOk = TFS_PATH_ERR;
                                    }
                                }
                            }
                        } else {
                            if(EntryPTR->Entry.Attr & TFS_ATTR_DIR) {
                                while(EntryPTR->Entry.Attr & TFS_ATTR_LINK) {
                                    EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Sub);
                                    if(!(EntryPTR) || (EntryPTR->LinkCnt == 0)) {
                                        return(TFS_NO_LINK);
                                    }
                                }
                                currPTR = (TFS_ENTRY OS_HUGE *)(EntryPTR);
                                Pos++;                     // over the '/' to next part of path
                            } else {
                                Pos = TFSstrlen(Name);
                                returnOk = TFS_PATH_ERR;
                            }
                        }
                    } else {                               // no entries in this DIR
                        Pos = TFSstrlen(Name);
                        returnOk = TFS_NOT_EXIST;
                    }
                } else {                                   // wrong element in PATH
                    Pos = TFSstrlen(Name);
                    returnOk = TFS_PATH_ERR;
                }
            } else {                                       // up to last element of PATH
                if((Pos + TFS_Name_SIZE) <= TFSstrlen(Name)) { // error in PATH
                    returnOk = TFS_PATH_ERR;
                }
                Pos = TFSstrlen(Name);
            }
        }
    }
    TFS_User[user]->currentEntry = currPTR;                // restore saved PATH / set actual PATH
    return(returnOk);
}

/*
************************************************************
*                      RESIZE ENTRY (internal)             *
************************************************************
*/
static U08 TFSResize_File(TFS_DIRFILE OS_HUGE *FilePTR, TFS_LONG size) KC51_REENTRANT
{
    TFS_DATA OS_HUGE    *DataPTR;
    U08                  returnOk;

    returnOk = TFS_NO_ERR;
    if(FilePTR->Sub != NULL) {
        DataPTR = (TFS_DATA OS_HUGE *)(FilePTR->Sub);
        while((size > DataPTR->Length) && (DataPTR->Entry.Next != NULL)) {
            size -= DataPTR->Length;
            DataPTR = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);
        }                                                  // calculate DB-Block as last
        if((size > DataPTR->Length) && (DataPTR->Entry.Next == NULL)) {
            returnOk = TFSExpand_File(FilePTR, size);
        } else {
            if((size <= DataPTR->Length) && (DataPTR->Entry.Next != NULL)) {
                while((DataPTR->Entry.Next != NULL) && (returnOk == TFS_NO_ERR)) {
#if (TFSDBLLINKED==1)
                    returnOk = TFSDelete_Entry(DataPTR->Entry.Next);
#else
                    returnOk = TFSDelete_Entry(DataPTR->Entry.Next, (TFS_ENTRY OS_HUGE *)(DataPTR));
#endif
                }
            }
        }
    } else {
        returnOk = TFSExpand_File(FilePTR, size);
    }
    if(returnOk == TFS_NO_ERR) {
        size = 0;
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&FilePTR->UsedLength), (U08 OS_HUGE *)(&size), sizeof(TFS_LONG));
    }
    return(returnOk);
}

/*
************************************************************
*             EXPAND/ADD DB-ENTRY (internal)               *
************************************************************
*/
static U08 TFSExpand_File(TFS_DIRFILE OS_HUGE *FilePTR, TFS_LONG size) KC51_REENTRANT
{
    TFS_ENTRY  OS_HUGE  *EntryPTR;
#if (!(TFSDBLLINKED==1))
    TFS_ENTRY  OS_HUGE  *LastPTR;
#endif
    TFS_DATA   OS_HUGE  *DataPTR;
    TFS_DATA             temp_DB;
    TFS_LONG             newsize;
    U08                  returnOk;

    newsize=0;
    if(size < TFS_AddMin) {
        size = TFS_AddMin;
    }
    if(FilePTR->Sub == NULL) {
        newsize  = size;
        returnOk = TFS_MEM_ERR;
#if (TFSDBLLINKED==1)
        temp_DB.Entry.Prev = (TFS_ENTRY OS_HUGE *)NULL;
#endif
    } else {
        DataPTR = (TFS_DATA OS_HUGE *)(FilePTR->Sub);
        while(DataPTR->Entry.Next != NULL) {
            DataPTR = (TFS_DATA OS_HUGE *)(DataPTR->Entry.Next);  // up to last DB-Block in this file
        }
        returnOk = TFSMem_Add((TFS_MEM OS_HUGE *)((U08 OS_HUGE *)(DataPTR) - sizeof(TFS_MEM)), size);
        if(returnOk == TFS_NO_ERR) {                       // get PTR to it and expand this MEM-Block
            newsize = DataPTR->Length + size;
            returnOk = TFSPort_Write((U08 OS_HUGE *)(&DataPTR->Length), (U08 OS_HUGE *)(&newsize), sizeof(TFS_LONG));
            newsize = 0;
        } else {                                           // change DB-Block size
            newsize = size;
        }
#if (TFSDBLLINKED==1)
        temp_DB.Entry.Prev = (TFS_ENTRY OS_HUGE *)(DataPTR);
#else
        LastPTR = (TFS_ENTRY OS_HUGE *)(DataPTR);
#endif
    }
    if(newsize > 0) {
        do {                                               // if can't last block expanded or file has no data
            returnOk = TFSMem_Alloc((TFS_ENTRY OS_HUGE **)(&EntryPTR), newsize + sizeof(TFS_DATA), &newsize);
            if(newsize) {                                  // add a new DB-Block
                newsize -= sizeof(TFS_DATA);
            }
            if(returnOk == TFS_NO_ERR) {
                temp_DB.Entry.Attr = TFS_ATTR_DB;
                temp_DB.Entry.PrevLevel = (TFS_ENTRY OS_HUGE *)(FilePTR);
                temp_DB.Entry.Next = (TFS_ENTRY OS_HUGE *)NULL;
                temp_DB.Length = newsize;
                returnOk = TFSPort_Write((U08 OS_HUGE *)(EntryPTR), (U08 OS_HUGE *)(&temp_DB), sizeof(TFS_DATA));
#if (TFSDBLLINKED==1)
                if((returnOk == TFS_NO_ERR) && (EntryPTR->Prev != NULL)) {
                    returnOk = TFSPort_Write((U08 OS_HUGE *)(&EntryPTR->Prev->Next), (U08 OS_HUGE *)(&EntryPTR), sizeof(U08 OS_HUGE *));
                }
                temp_DB.Entry.Prev = EntryPTR;
#else
                if((returnOk == TFS_NO_ERR) && (LastPTR != NULL)) {
                    returnOk = TFSPort_Write((U08 OS_HUGE *)(&LastPTR->Next), (U08 OS_HUGE *)(&EntryPTR), sizeof(U08 OS_HUGE *));
                }
                LastPTR = EntryPTR;
#endif
                if(newsize == size) {
                    size = newsize = 0;                    // handle fragmentation
                } else {
                    newsize = size;
                }
            } else {
                if(newsize != 0) {
                    size -= newsize;
                    if(EntryPTR->Next != NULL) {
                        EntryPTR = EntryPTR->Next;
                    }
                } else {
                    return(returnOk);
                }
            }                                              // notfragmented or framented
        } while((size > 0) || (newsize != 0));
    }
    return(returnOk);
}

/*
************************************************************
*         CREATE DIR/FILE incl. DBs ENTRY (internal)       *
************************************************************
*/
static U08 TFSCreate_Entry(TFS_PATHNAME OS_HUGE *Name, TFS_ATTR Attribs, TFS_LONG size) KC51_REENTRANT
{
    TFS_DIRFILE OS_HUGE *EntryPTR;
    TFS_DIRFILE OS_HUGE *DirPTR;
#if (!(TFSDBLLINKED==1))
    TFS_ENTRY   OS_HUGE *LastPTR;
#endif
    TFS_DIRFILE          temp_DF;
    TFS_LONG             biggest;
    U08                  returnOk, i, user;

    user = TFSGetUserByID();
    if(user == 0xFF) {
        return(TFS_NO_USER);
    }
    DirPTR = (TFS_DIRFILE OS_HUGE *)(TFS_User[user]->currentEntry); // in currcent DIR
    if(DirPTR->Sub != NULL) {
        EntryPTR = (TFS_DIRFILE OS_HUGE *)(DirPTR->Sub);   // get first entry and check if name exist
        while((0 != (i = TFSstrcmp(Name, (TFS_PATHNAME OS_HUGE *)EntryPTR->Name))) && (NULL != EntryPTR->Entry.Next)) {
            EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
        }
        if(i == 0) {                                       // up to end of level
            return(TFS_NAME_EXIST);                        // Entry-name exist
        }
    }
#if (TFSDBLLINKED==1)
    else {
        temp_DF.Entry.Prev = (TFS_ENTRY *)NULL;           // it's the first entry in this level
    }
#endif
    temp_DF.Entry.Attr = Attribs;
    temp_DF.Entry.PrevLevel = (TFS_ENTRY OS_HUGE *)(DirPTR);
    temp_DF.Entry.Next = (TFS_ENTRY OS_HUGE *)NULL;        // no entry after this
    temp_DF.LinkCnt = 0;                                   // no links to this entry
    temp_DF.UsedLength = 0;
    temp_DF.Sub = (TFS_ENTRY OS_HUGE *)NULL;               // no sub-elements
    biggest=size;
    do {
        if(DirPTR->Sub != NULL)                            // add a new entry
#if (TFSDBLLINKED==1)
            temp_DF.Entry.Prev = (TFS_ENTRY OS_HUGE *)(EntryPTR);
#else
            LastPTR = (TFS_ENTRY OS_HUGE *)(EntryPTR);
#endif
        if(Name != NULL) {
            returnOk = TFSMem_Alloc((TFS_ENTRY OS_HUGE **)(&EntryPTR), sizeof(TFS_DIRFILE), &biggest);
        } else {
            returnOk = TFSMem_Alloc((TFS_ENTRY OS_HUGE **)(&EntryPTR), biggest + sizeof(TFS_DATA), &biggest);
        }
        if(biggest) {
            biggest -= sizeof(TFS_DATA);                   // handle fragmentation for DB-Blocks
        }
        if(returnOk == TFS_NO_ERR) {
            if(Name != NULL) {                             // if DIR or FILE -Header
                i = 0;
                for(i=0; (i < TFSstrlen(Name)) && (i < (TFS_Name_SIZE-1)); i++) {
                    temp_DF.Name[i] = Name[i];
                }
                for(;i < TFS_Name_SIZE; i++) {
                    temp_DF.Name[i] = 0;
                }
                returnOk = TFSPort_Write((U08 OS_HUGE *)(EntryPTR), (U08 OS_HUGE *)(&temp_DF), sizeof(TFS_DIRFILE));
            } else {                                       // if DB-Block
                returnOk = TFSPort_Write((U08 OS_HUGE *)(EntryPTR), (U08 OS_HUGE *)(&temp_DF), sizeof(TFS_ENTRY));
                if(returnOk == TFS_NO_ERR) {
                    returnOk = TFSPort_Write((U08 OS_HUGE *)(EntryPTR) + sizeof(TFS_ENTRY), (U08 OS_HUGE *)(&biggest), sizeof(TFS_LONG));
                }
                if(biggest == size) {                      // write lenght in DB
                    size = biggest = 0;
                } else {
                    biggest = size;                        // handle fragmentation
                }
            }
            if(DirPTR->Sub == NULL) {                      // update hirarchical pointer-entries
                returnOk = TFSPort_Write((U08 OS_HUGE *)(&DirPTR->Sub), (U08 OS_HUGE *)(&EntryPTR), sizeof(U08 OS_HUGE *));
            } else {
#if (TFSDBLLINKED==1)
                returnOk = TFSPort_Write((U08 OS_HUGE *)(&EntryPTR->Entry.Prev->Next), (U08 OS_HUGE *)(&EntryPTR), sizeof(U08 OS_HUGE *));
#else
                returnOk = TFSPort_Write((U08 OS_HUGE *)(&LastPTR->Next), (U08 OS_HUGE *)(&EntryPTR), sizeof(U08 OS_HUGE *));
#endif
            }
        } else {                                           // handle fragmentation (only for DB-Blocks)
            if((Name == NULL) && (biggest != 0)) {
                size -= biggest;
                if(EntryPTR->Entry.Next != NULL)
                    EntryPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->Entry.Next);
            } else {                                       // if DIR or FILE entry - abort (no fragmented header)
                return(returnOk);
            }
        }                                                  // up to end of fragmentation
    } while((Name == NULL) && ((size > 0) || (biggest != 0)));
    return(returnOk);
}

/*
************************************************************
*          UNLINK DIR/FILE or DB ENTRY (internal)          *
************************************************************
*/
#if (TFSDBLLINKED==1)
 static U08 TFSUnlink_Entry(TFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT
#else
 static U08 TFSUnlink_Entry(TFS_ENTRY OS_HUGE *EntryPTR, TFS_ENTRY OS_HUGE *PrevPTR) KC51_REENTRANT
#endif
{
    TFS_DIRFILE OS_HUGE  *DirPTR;
    TFS_ENTRY   OS_HUGE  *tmpPTR;
    U08                   returnOk;

    DirPTR = (TFS_DIRFILE OS_HUGE *)(EntryPTR->PrevLevel);
    tmpPTR = EntryPTR->Next;
#if (TFSDBLLINKED==1)
    if(EntryPTR->Prev == NULL) {                           // update hirarchical pointer-system
#else
    if(PrevPTR == NULL) {
#endif
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&DirPTR->Sub), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
    } else {
#if (TFSDBLLINKED==1)
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&EntryPTR->Prev->Next), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
    }
    if((EntryPTR->Next != NULL) && (returnOk == TFS_NO_ERR)) {
        tmpPTR = EntryPTR->Prev;
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&EntryPTR->Next->Prev), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
    }
#else
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&PrevPTR->Next), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
    }
#endif
    return(returnOk);
}

/*
************************************************************
*          DELETE DIR/FILE or DB ENTRY (internal)          *
************************************************************
*/
#if (TFSDBLLINKED==1)
 static U08 TFSDelete_Entry(TFS_ENTRY OS_HUGE *EntryPTR) KC51_REENTRANT
#else
 static U08 TFSDelete_Entry(TFS_ENTRY OS_HUGE *EntryPTR, TFS_ENTRY OS_HUGE *PrevPTR) KC51_REENTRANT
#endif
{
    U08   returnOk;

#if (TFSDBLLINKED==1)
    returnOk = TFSUnlink_Entry(EntryPTR);
#else
    returnOk = TFSUnlink_Entry(EntryPTR, PrevPTR);
#endif
    if(returnOk == TFS_NO_ERR) {                           // clear entry
        returnOk = TFSPort_Fill((U08 OS_HUGE *)(EntryPTR), 0, sizeof(TFS_DIRFILE));
        if(returnOk == TFS_NO_ERR) {                       // free Memory (only in MEM-list)
            returnOk = TFSMem_Free((TFS_MEM OS_HUGE *)(EntryPTR));
        }
    }
    return(returnOk);
}

/*
************************************************************
*                 FORMAT TFS-MEMORY (internal)             *
************************************************************
*/
static U08 TFSMem_Format(void) KC51_REENTRANT
{
    TFS_MEM OS_HUGE   *EntryPTR;
    TFS_MEM            temp_MPTR;
    U08                returnOk;

    returnOk = TFSPort_Fill((U08 OS_HUGE *)(TFS_START), 0x00, (TFS_LONG)TFS_SIZE);
    if(returnOk == TFS_NO_ERR) {
        temp_MPTR.Next = (TFS_MEM OS_HUGE *)((U08 OS_HUGE *)(TFS_START)+TFS_SIZE-sizeof(TFS_MEM));
#if (TFSDBLLINKED==1)
        temp_MPTR.Prev = (TFS_MEM OS_HUGE *)NULL;          // setup the startup-entry's at begin ...
#endif
        temp_MPTR.Len = 0;
        returnOk = TFSPort_Write((U08 OS_HUGE *)(TFS_START), (U08 OS_HUGE *)(&temp_MPTR), sizeof(TFS_MEM));
        if(returnOk == TFS_NO_ERR) {
            EntryPTR = temp_MPTR.Next;
            temp_MPTR.Next = (TFS_MEM OS_HUGE *)NULL;
#if (TFSDBLLINKED==1)
            temp_MPTR.Prev = (TFS_MEM OS_HUGE *)(TFS_START); // ... and end
#endif
            temp_MPTR.Len = 0;
            returnOk = TFSPort_Write((U08 OS_HUGE *)(EntryPTR), (U08 OS_HUGE *)(&temp_MPTR), sizeof(TFS_MEM));
        }
    }
    return(returnOk);
}

/*
************************************************************
*         GET FREE-SIZE OF BRUTTO-MEMORY (internal)        *
************************************************************
*/
static TFS_LONG TFSMem_FreeSize(void) KC51_REENTRANT
{
    TFS_MEM OS_HUGE *cur_TFSM;
    TFS_LONG         gsize, length;

    cur_TFSM = (TFS_MEM OS_HUGE *)(TFS_START);
    gsize = 0;                                             // from begin
    while(1) {                                             // search free space
        length = (TFS_LONG)((U08 OS_HUGE *)(cur_TFSM->Next) - (U08 OS_HUGE *)(cur_TFSM)) - cur_TFSM->Len;
        if(length > sizeof(TFS_MEM)) {
            gsize += (length-sizeof(TFS_MEM));
        }
        cur_TFSM = cur_TFSM->Next;                         // calculate free-size of drive
        if(!(cur_TFSM->Next)) {                            // up to end of drive
            break;
        }
    }
    return(gsize);                                         // return brutto-size (fragmentation is not checked)
}

/*
************************************************************
*                 ALLOCATE TFS-MEMORY (internal)           *
************************************************************
*/
static U08 TFSMem_Alloc(TFS_ENTRY OS_HUGE *(*TFSMPtr), TFS_LONG size, TFS_LONG *biggest) KC51_REENTRANT
{
    TFS_MEM OS_HUGE *cur_TFSM;
    TFS_MEM OS_HUGE *new_TFSM;
    TFS_MEM          temp_TFSM;
    TFS_LONG         gsize, length;
    U08              returnOk;

    if((size == 0) || (size > (((TFS_LONG)TFS_SIZE) - (2*sizeof(TFS_MEM) + sizeof(TFS_DIRFILE))))) {
        return(TFS_MEM_ERR);                               // check if size invalid or greater the size of drive
    }
    size += sizeof(TFS_MEM);                               // add header
    cur_TFSM = (TFS_MEM OS_HUGE *)(TFS_START);
    gsize = 0;                                             // from begin
    *biggest = 0;
    while(1) {                                             // search free space
        length = (TFS_LONG)((U08 OS_HUGE *)(cur_TFSM->Next) - (U08 OS_HUGE *)(cur_TFSM)) - cur_TFSM->Len;
        if(length >= size) {
            break;                                         // break if found
        }
        if(*biggest < length) {
            *biggest = length;                             // save greatest block
        }
        if(length > (sizeof(TFS_MEM) + sizeof(TFS_DATA))) {
            gsize += (length-sizeof(TFS_MEM) - sizeof(TFS_DATA));
        }
        cur_TFSM = cur_TFSM->Next;                         // calculate free-size of drive
        if(!(cur_TFSM->Next)) {                            // up to end of drive
            if(gsize >= (size - sizeof(TFS_MEM))) {
                *biggest -= sizeof(TFS_MEM);
            } else {                                       // check size of drive to needed size
                *biggest = 0;
            }
            return(TFS_MEM_OVF);                           // not enough memory (for this block / in this drive)
        }
    }
    *biggest = size - sizeof(TFS_MEM);
    if(!(cur_TFSM->Len)) {                                 // handle first allocation
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&cur_TFSM->Len), (U08 OS_HUGE *)(&size), sizeof(TFS_LONG));
        if(returnOk == TFS_NO_ERR) {
            *TFSMPtr = (TFS_ENTRY OS_HUGE *)((U08 OS_HUGE *)(cur_TFSM) + sizeof(TFS_MEM));
        }
        return(returnOk);
    }                                                      // handle all later allocations
    new_TFSM = (TFS_MEM OS_HUGE *)((U08 OS_HUGE *)(cur_TFSM)+cur_TFSM->Len);
    temp_TFSM.Next = cur_TFSM->Next;
#if (TFSDBLLINKED==1)
    temp_TFSM.Prev = cur_TFSM;
#endif
    temp_TFSM.Len = size;                                  // create new block and update pointer-system
    returnOk = TFSPort_Write((U08 OS_HUGE *)(new_TFSM), (U08 OS_HUGE *)(&temp_TFSM), sizeof(TFS_MEM));
    if(returnOk == TFS_NO_ERR) {
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&cur_TFSM->Next), (U08 OS_HUGE *)(&new_TFSM), sizeof(U08 OS_HUGE *));
    }
#if (TFSDBLLINKED==1)
    if((new_TFSM->Next) && (returnOk == TFS_NO_ERR)) {
        returnOk = TFSPort_Write((U08 OS_HUGE *)(&new_TFSM->Next->Prev), (U08 OS_HUGE *)(&new_TFSM), sizeof(U08 OS_HUGE *));
    }
#endif
    if(returnOk == TFS_NO_ERR) {
        *TFSMPtr = (TFS_ENTRY OS_HUGE *)((U08 OS_HUGE *)(new_TFSM) + sizeof(TFS_MEM));
    }
    return(returnOk);
}

/*
************************************************************
*                  ADD TFS-MEMORY (internal)               *
************************************************************
*/
static U08 TFSMem_Add(TFS_MEM OS_HUGE *TFSMPtr, TFS_LONG size) KC51_REENTRANT
{
    if(!(size) || (size > ((TFS_LONG)(0-1) - sizeof(TFS_MEM)))) { // ckeck size
        return(TFS_MEM_ERR);
    }
#if (TFSDBLLINKED==1)
    if((TFSMPtr == (TFS_MEM OS_HUGE *)(TFS_START)) || (TFSMPtr == (TFSMPtr->Prev->Next))) {
#endif
        if((((U08 OS_HUGE *)(TFSMPtr->Next)) - ((U08 OS_HUGE *)(TFSMPtr)) - TFSMPtr->Len) >= size) {
            size += TFSMPtr->Len;                          // if enough free-MEM after this block add needed size
            return(TFSPort_Write((U08 OS_HUGE *)(&TFSMPtr->Len), (U08 OS_HUGE *)(&size), sizeof(TFS_LONG)));
        } else {
            return(TFS_MEM_OVF);                           // no or not enough free memory after this block
        }
#if (TFSDBLLINKED==1)
    }
    return(TFS_MEM_ERR);                                   // something wrong in memory-system (pointer)
#endif
}

/*
************************************************************
*                  FREE TFS-MEMORY (internal)              *
************************************************************
*/
static U08 TFSMem_Free(TFS_MEM OS_HUGE *TFSMPtr) KC51_REENTRANT
{
    TFS_MEM OS_HUGE   *tmpPTR;
#if (!(TFSDBLLINKED==1))
    TFS_MEM OS_HUGE   *PrevMEM;
#endif
    U08                returnOk;

    if(!(TFSMPtr)) {
        return(TFS_MEM_ERR);
    }
    TFSMPtr = (TFS_MEM OS_HUGE *)((U08 OS_HUGE *)(TFSMPtr)-sizeof(TFS_MEM));
#if (TFSDBLLINKED==1)
    if((TFSMPtr == (TFS_MEM OS_HUGE *)(TFS_START)) || (TFSMPtr == (TFSMPtr->Prev->Next))) {
        if((TFSMPtr->Prev)) {                              // check if the pointer is in correct chain
            tmpPTR = TFSMPtr->Next;                        // update pointer..
            returnOk = TFSPort_Write((U08 OS_HUGE *)(&TFSMPtr->Prev->Next), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
            if((TFSMPtr->Next) && (returnOk == TFS_NO_ERR)) {
                tmpPTR = TFSMPtr->Prev;
                returnOk = TFSPort_Write((U08 OS_HUGE *)(&TFSMPtr->Next->Prev), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
            }
            if(returnOk == TFS_NO_ERR) {                   // and clear MEM-header
                TFSPort_Fill((U08 OS_HUGE *)(TFSMPtr), 0, sizeof(TFS_MEM));
            }
            return(returnOk);
        } else {                                           // if was only this block allocated - clear lenght in start-entry
            TFSPort_Fill((U08 OS_HUGE *)(&TFSMPtr->Len), 0, sizeof(TFS_LONG));
        }
    } else {
        return(TFS_MEM_ERR);                               // something wrong in memory-system (pointer)
    }
#else
    tmpPTR = (TFS_MEM OS_HUGE *)(TFS_START);
    PrevMEM = (TFS_MEM OS_HUGE *)NULL;
    while((tmpPTR != TFSMPtr) && (tmpPTR != NULL)) {
        PrevMEM = tmpPTR;                                  // save prev-pointer
        tmpPTR = tmpPTR->Next;                             // ..goto next.. upto end or correct position
    }
    if(tmpPTR != NULL) {
        if(PrevMEM != NULL) {
            tmpPTR = TFSMPtr->Next;
            returnOk = TFSPort_Write((U08 OS_HUGE *)(&PrevMEM->Next), (U08 OS_HUGE *)(&tmpPTR), sizeof(U08 OS_HUGE *));
            if(returnOk == TFS_NO_ERR) {                   // and clear MEM-header
                TFSPort_Fill((U08 OS_HUGE *)(TFSMPtr), 0, sizeof(TFS_MEM));
            }
        } else {
            TFSPort_Fill((U08 OS_HUGE *)(&TFSMPtr->Len), 0, sizeof(TFS_LONG));
        }
    } else {
        return(TFS_MEM_ERR);                               // something wrong in memory-system (pointer)
    }
#endif
    return(TFS_NO_ERR);
}

/************************* END ****************************/


